DEFINT A-Z

'$INCLUDE: '$YME.BI'
'$INCLUDE: 'C:\UTILS\QB45\BAS\QB.BI'

CONST scINDEX = &H3C4             ' Sequence Controller Index Register's port
CONST scDATA = &H3C5              ' SC Data Register's port
CONST scMapMask = 2               ' SC Register Index
CONST scSequencerMemoryMode = 4   ' SC Register Index
CONST gcINDEX = &H3CE             ' Graphic Controller Index Register's port
CONST gcDATA = &H3CF              ' GC Data Register's port
CONST gcReadMapSelect = 4         ' GC Register Index
CONST gcBitMask = 8               ' GC Register Index
CONST crtcINDEX2 = &H3D4          ' CRT Controller Index Register's port
CONST crtcDATA2 = &H3D5           ' CRTC Data Register's port
CONST crtcStartAddressHigh = 12   ' CRTC Register Index
CONST crtcUnderlineLocation = 20  ' CRTC Register Index
CONST crtcModeControl = 23        ' CRTC Register Index

CONST dacPelMask = &H3C6          ' Digital/Analog Converter PEL Mask register
CONST dacREAD = &H3C7             ' DAC Read Port/DAC State Register
CONST dacWRITE = &H3C8            ' DAC Write Port
CONST dacDATA = &H3C9             ' DAC Data Port

CONST pg0sg = &HA000              ' Double-buffering (front)
CONST pg1sg = &HA3E8              ' Sprite bank
CONST pg2sg = &HA7D0              ' Double-buffering (back)
CONST pg3sg = &HABB8              ' Solid bank

CONST maxBankSize = 256           ' Half the image array

TYPE imgType            ' OFS SZE  DESCRIPTION
	ofs AS INTEGER        ' 000 002  Offset to the first pixel
	wr AS INTEGER         ' 002 002  Offset for the writing cursor (relative)
	w AS INTEGER          ' 004 002  Width in pixels
	h AS INTEGER          ' 006 002  Height in pixels
	w2 AS INTEGER         ' 008 002  Width rounded up to the next multiple of 4
	size AS INTEGER       ' 010 002  Size of the buffer, in bytes (w2 * h)
	x AS INTEGER          ' 012 002  Horizontal display offset
	y AS INTEGER          ' 014 002  Vertical display offset
END TYPE                ' 16 BYTES - IMAGE DESCRIPTOR (SPRITE & SOLID)

type fontType						' OFS SZE  DESCRIPTION
	imgId as integer			' 000 002  Index of the reference image
	ofs as integer				' 002 002  Offset to first pixel (in sprite sheet)
	w as integer					' 004 002  4 (4x8) or 8 (8x8)
	first as integer			' 006 002  First ASCII character
	last as integer				' 008 002  Last ASCII character (at most 128!)
	maskSeg as integer		' 010 002  Segment to mask data
	padding as long				' 012 004  /
END TYPE								' 16 BYTES - CHARACTER MAP DESCRIPTOR

TYPE vgaType            ' OFS SZE  DESCRIPTION
	workHiByte AS INTEGER ' 000 002  Working page offset high byte
	workMemSeg AS INTEGER ' 002 002  Working page full memory segment
	viewHiByte AS INTEGER ' 004 002  Display page offset high byte
	viewMemSeg AS INTEGER ' 006 002  Display page full memory segment
	viewportX1 AS INTEGER ' 008 002  Viewport left edge
	viewportY1 AS INTEGER ' 010 002  Viewport top edge
	viewportX2 AS INTEGER ' 012 002  Viewport right edge
	viewportY2 AS INTEGER ' 014 002  Viewport bottom edge
	viewportW AS INTEGER  ' 016 002  Viewport width, in pixels
	viewportH AS INTEGER  ' 018 002  Viewport height, in pixels
	clearAs AS INTEGER    ' 020 002  0: leave, 1: zero, 2: copy frame
	spriteOfs AS INTEGER  ' 022 002  Writing offset in bank 1
	solidOfs AS INTEGER   ' 024 002  Writing offset in bank 2
END TYPE                ' 26 BYTES

TYPE palType						' OFS SZE  DESCRIPTION
	clr AS STRING * 768		' 000 768
	inUse AS INTEGER			' 768 002
END TYPE 								' 770 BYTES

DECLARE SUB memCopy ALIAS "B$ASSN" (BYVAL FSeg%, BYVAL FOfs%, BYVAL Bytes1%, BYVAL TSeg%, BYVAL tOfs%, BYVAL Bytes2%)

DIM SHARED image(511) AS imgType, font(15) AS fontType, clrMap(3) AS palType
DIM SHARED iSqr(7) AS INTEGER, vga AS vgaType

''
'' Copy {count} bytes of memory from srcSeg:srcOfs to dstSeg:dstOfs.
'' B$ASSN is a compiler-only instruction and is not available in QuickBASIC.
'' Using the simple DECLARE is not enough to have a program use it outside the
'' library, hence this dirty routine.
''
SUB memCpy(srcSeg as integer, srcOfs as integer, dstSeg as integer, dstOfs as integer, count as integer)
	if (count) then memCopy srcSeg, srcOfs, count, dstSeg, dstOfs, count
end SUB

''
'' Test for existing file (doesn't check for archives!) (DOS 2+)
''
FUNCTION fileExists% (flName AS STRING) STATIC
	DIM DTA AS STRING * 44, regs AS RegTypeX, flNameZ AS STRING

	flNameZ = flName + CHR$(0)

	'' Set up our own Disk Transfer Area (DTA)
	regs.ax = &H1A00                    ' DOS 1+ - Set Disk Transfer Area Address
	regs.DX = VARPTR(DTA)               ' Memory offset to DTA
	regs.ds = VARSEG(DTA)               ' Memory segment to DTA
	CALL INTERRUPTX(&H21, regs, regs)   ' Do

	'' Search for file
	regs.ax = &H4E00                    ' DOS 2+ - FindFirst - Find First Matching File
	regs.CX = 0                         ' Object type (0: file only)
	regs.DX = SADD(flNameZ)             ' Memory offset to flNameZ
	regs.ds = VARSEG(flNameZ)           ' Memory segment to flNameZ
	CALL INTERRUPTX(&H21, regs, regs)   ' Do

	'' Returns TRUE if file exists
	fileExists% = ((regs.flags AND &H1) = 0)
END FUNCTION

''
'' ALLOCATE MEMORY (DOS 2+)
''
FUNCTION memAlloc% (numBytes AS LONG)
	DIM memReq AS INTEGER, junk AS LONG, regs AS RegTypeX

	' Paragraphs are groups of 16 bytes
	memReq = (numBytes \ 16) - ((numBytes AND 15) > 0)
	IF (memReq = 0) THEN EXIT FUNCTION

	' Tell QuickBASIC to free some memory (not sure why a margin is needed)
	junk = SETMEM(-CLNG(memReq + 1) * 16)

	' Use DOS Interrupt 0x48 to request <memReq> paragraphs of memory
	regs.ax = &H4800
	regs.bx = memReq
	CALL INTERRUPTX(&H21, regs, regs)

	' If CF is not clear, something went wrong
	IF (regs.flags AND &H1) THEN
		junk = SETMEM(650000)
		memAlloc% = 0
	ELSE
		memAlloc% = regs.ax
	END IF
END FUNCTION

''
'' FREE MEMORY ALLOCATED VIA DOS INTERRUPT 0x21, FUNCTION 0x48
''
SUB memFree (segAdr AS INTEGER)
	DIM junk AS LONG, regs AS RegTypeX

	' No segment specified, abort
	IF (segAdr = 0) THEN EXIT SUB

	' Free allocated memory
	regs.ax = &HA900
	regs.es = segAdr
	CALL INTERRUPTX(&H21, regs, regs)

	' Clear segment and offset
	segAdr = 0

	' Give back memory to QuickBASIC
	junk = SETMEM(650000)
END SUB

''
'' save view page to bitmap file
''
'SUB yCapture (filename AS STRING, pg as integer)
SUB yCapture (filename AS STRING)
	DIM temp AS LONG, scan(159) AS LONG, dstSeg AS INTEGER
	DIM ff AS INTEGER, srcOfs AS INTEGER
	DIM scanStr AS STRING * 320

	DIM bfType AS INTEGER, bfSize AS LONG
	DIM bfReserved AS LONG, bfOffByte AS LONG
	DIM biSize AS LONG, biWidth AS LONG, biHeight AS LONG
	DIM biPlanes AS INTEGER, biBitCount AS INTEGER
	DIM biCompression AS LONG, biSizeImage AS LONG
	DIM biXPelsPerMeter AS LONG, biYPelsPerMeter AS LONG
	DIM biClrUsed AS LONG, biClrImportant AS LONG

	' select plane read mode
	OUT gcINDEX, gcReadMapSelect

	' get file handle
	ff = FREEFILE
	IF (ff) THEN
		OPEN filename FOR BINARY AS #ff

		' write header
		bfType = &H4D42     ' "BM" - type of file
		bfSize = 65080      ' Size of the whole file, in bytes
		'bfReserved = 0     ' Reseved, always null
		bfOffByte = 1078    ' Start of bitmap data
		PUT #ff, , bfType
		PUT #ff, , bfSize
		PUT #ff, , bfReserved
		PUT #ff, , bfOffByte

		biSize = 40         ' Size of header, in bytes
		biWidth = 320       ' Width, in pixels
		biHeight = 200      ' Height, in pixels
		biPlanes = 1        ' Number of planes
		biBitCount = 8      ' Number of bits per pixel
		'biCompression = 0  ' Compression method
		biSizeImage = 64000 ' Size of the whole image, in bytes
		'biXPelsPerMeter = 0' Horizontal resolution, in pixels per meter
		'biYPelsPerMeter = 0' Horizontal resolution, in pixels per meter
		'biClrUsed = 0      ' Colors used (0 for all)
		'biClrImportant = 0 ' First important colors (0 for all)
		PUT #ff, , biSize
		PUT #ff, , biWidth
		PUT #ff, , biHeight
		PUT #ff, , biPlanes
		PUT #ff, , biBitCount
		PUT #ff, , biCompression
		PUT #ff, , biSizeImage
		PUT #ff, , biXPelsPerMeter
		PUT #ff, , biYPelsPerMeter
		PUT #ff, , biClrUsed
		PUT #ff, , biClrImportant

		' color map
		OUT dacREAD, 0
		FOR i% = 0 TO 255
			temp = (INP(dacDATA) * 65536 + INP(dacDATA) * 256 + INP(dacDATA)) * 4
			PUT #ff, , temp
		NEXT i%

		' start at line 199
		srcOfs = 15920
		dstSeg = VARSEG(scan(0))
		DEF SEG = dstSeg
		DO
			' read planar scanline
			OUT gcDATA, 0: memCopy vga.viewMemSeg, srcOfs, 80, dstSeg, 0, 80
			OUT gcDATA, 1: memCopy vga.viewMemSeg, srcOfs, 80, dstSeg, 80, 80
			OUT gcDATA, 2: memCopy vga.viewMemSeg, srcOfs, 80, dstSeg, 160, 80
			OUT gcDATA, 3: memCopy vga.viewMemSeg, srcOfs, 80, dstSeg, 240, 80
			'OUT gcDATA, 0: memCopy &hA000 + (pg * 1000), srcOfs, 80, dstSeg, 0, 80
			'OUT gcDATA, 1: memCopy &hA000 + (pg * 1000), srcOfs, 80, dstSeg, 80, 80
			'OUT gcDATA, 2: memCopy &hA000 + (pg * 1000), srcOfs, 80, dstSeg, 160, 80
			'OUT gcDATA, 3: memCopy &hA000 + (pg * 1000), srcOfs, 80, dstSeg, 240, 80

			' convert planar to linear
			tmp = 320
			FOR i% = 0 TO 79
				POKE tmp, PEEK(i%)
				POKE tmp + 1, PEEK(i% + 80)
				POKE tmp + 2, PEEK(i% + 160)
				POKE tmp + 3, PEEK(i% + 240)
				tmp = tmp + 4
			NEXT i%

			' write linear to file
			memCopy dstSeg, 320, 320, VARSEG(scanStr), VARPTR(scanStr), 320
			PUT #ff, , scanStr

			' move to scanline above
			srcOfs = srcOfs - 80
		LOOP UNTIL (srcOfs < 0)

		CLOSE #ff
	END IF
	OUT gcINDEX, gcBitMask
END SUB

''
'' clear character map
''
SUB yCMapFree (id AS INTEGER)
	' stay in range
	SELECT CASE id
	CASE LBOUND(font) TO UBOUND(font)
		memFree font(id).maskSeg
		font(id).w = 0
	END SELECT
END SUB

''
'' SETUP CHARACTER MAP (width is either 4 or 8, height is always 8)
''
FUNCTION yCMapSet% (imgId AS INTEGER, isSprite as integer, first AS INTEGER, last AS INTEGER)
	DIM mask AS STRING, tmpOfs as integer

	' assume failure to initialize
	yCMapSet% = -1

	' stay in range
	SELECT CASE imgId
	CASE IS < 0, IS >= maxBankSize
		EXIT FUNCTION
	END SELECT

	' keep ASCII codes in check
	SELECT CASE first
	CASE IS < 0, IS > last
		EXIT FUNCTION
	END SELECT
	SELECT CASE last
	CASE IS < first, IS > 255
		EXIT FUNCTION
	END SELECT

	' check dimensions (image must be 4 or 8 wide, characters are 8 pixels tall)
	if ((image(imgId).w <> 4) AND (image(imgId).w <> 8)) then exit function
	IF (image(imgId).h <> ((last - first + 1) * 8)) THEN EXIT FUNCTION

	' find a free character map
	FOR i% = LBOUND(font) TO UBOUND(font)
		IF (font(i%).w = 0) THEN
			' memory offset and base image
			font(i%).ofs = image(imgId).ofs
			font(i%).imgId = imgId
			' dimensions
			font(i%).w = image(imgId).w
			' range
			font(i%).first = first
			font(i%).last = last
			' transparency mask if necessary
			if (isSprite) then
				mask = string$(image(imgId).w * (last - first + 1), 0)
				font(i%).maskSeg = memAlloc%(len(mask))
				if (font(i%).maskSeg) then
					' switch to read mode
					OUT gcINDEX, gcReadMapSelect
					' for each plane, rebuild the mask (tag opaque pixels)
					DEF SEG = pg1sg
					tmpOfs = image(imgId).ofs
					for p% = 0 to 3
						OUT gcDATA, p%
						for j% = 0 to (image(imgId).size \ 4) - 1 step 2
							if (peek(tmpOfs + j%) < 255) then
								if (peek(tmpOfs + j% + 1) < 255) then
									mid$(mask, 1 + j%\2, 1) = chr$(asc(mid$(mask, 1 + j%\2,1)) OR iSqr(p%) OR iSqr(p% + 4))
								else
									mid$(mask, 1 + j%\2, 1) = chr$(asc(mid$(mask, 1 + j%\2,1)) OR iSqr(p%))
								end if
							elseif (peek(tmpOfs + j% + 1) < 255) then
								mid$(mask, 1 + j%\2, 1) = chr$(asc(mid$(mask, 1 + j%\2,1)) OR iSqr(p% + 4))
							end if
						next j%
					next p%
					' switch back to mask mode
					OUT gcINDEX, gcBitMask
					' preserve mask
					memCopy varseg(mask), sadd(mask),len(mask) , font(i%).maskSeg, 0, len(mask)
				end if
			else
				font(i%).maskSeg = 0
			end if
			' and we're done
			yCMapSet% = i%
			EXIT FOR
		END IF
	NEXT i%
END FUNCTION

''
'' FLIP BACK AND FRONT PAGES
''
SUB yFlip
	' swap work and display page data
	SWAP vga.workHiByte, vga.viewHiByte ' high byte
	SWAP vga.workMemSeg, vga.viewMemSeg ' segment

	' replace starting address (high byte only)
	DO: LOOP WHILE (INP(&H3DA) AND 1)
	OUT crtcDATA2, vga.viewHiByte
	DO: LOOP UNTIL (INP(&H3DA) AND 8)
END SUB

''
'' DRAW IMAGE IN VIEWPORT
''
SUB yImgDraw (id AS INTEGER, x AS INTEGER, y AS INTEGER)
	DIM ptX1 AS INTEGER, srcW AS INTEGER
	DIM ptY1 AS INTEGER, srcH AS INTEGER

	DIM isSprite AS INTEGER, lastFullPlane AS INTEGER, pxl AS INTEGER

	DIM srcOfs AS INTEGER, srcPln AS INTEGER, ofs1 AS INTEGER
	DIM dstOfs AS INTEGER, dstPln AS INTEGER, ofs2 AS INTEGER
	DIM stride AS INTEGER, group4 AS INTEGER

	SELECT CASE id
	' out of range, bail
	CASE IS < LBOUND(image), IS > UBOUND(image)
		EXIT SUB
	' sprite
	CASE IS < maxBankSize
		isSprite = -1
	' solid
	CASE ELSE
		isSprite = 0
	END SELECT

	' starting offset in source buffer
	srcOfs = image(id).ofs

	'' ---- horizontal boundaries ---- ''
	ptX1 = x - image(id).x
	srcW = image(id).w

	' crop left
	SELECT CASE ptX1
	CASE IS > vga.viewportW - 1
		EXIT SUB
	CASE IS < 0
		srcOfs = srcOfs + ((-ptX1) \ 4) ' start further right
		srcPln = ((-ptX1) AND &H3)
		srcW = srcW + ptX1
		ptX1 = 0
	CASE ELSE
		srcPln = 0
	END SELECT

	' crop right, get width
	SELECT CASE ptX1 + srcW - 1
	CASE IS > vga.viewportW - 1
		srcW = vga.viewportW - ptX1
	CASE IS < 0
		EXIT SUB
	END SELECT

	' adjust coordinates to viewport
	ptX1 = ptX1 + vga.viewportX1

	'' ---- vertical boundaries ---- ''
	ptY1 = y - image(id).y
	srcH = image(id).h

	' crop top
	SELECT CASE ptY1
	CASE IS > vga.viewportH - 1
		EXIT SUB
	CASE IS < 0
		srcOfs = srcOfs + ((image(id).w2 \ 4) * (-ptY1)) ' start further down
		srcH = srcH + ptY1
		ptY1 = 0
	END SELECT

	' crop bottom, get height
	SELECT CASE ptY1 + srcH - 1
	CASE IS > vga.viewportH - 1
		srcH = vga.viewportH - ptY1
	CASE IS < 0
		EXIT SUB
	END SELECT

	' adjust coordinates to viewport
	ptY1 = ptY1 + vga.viewportY1

	IF ((srcH < 1) OR (srcW < 1)) THEN EXIT SUB

	' target plane
	dstOfs = (ptX1 \ 4) + (ptY1 * 80)
	dstPln = (ptX1 AND &H3)
	stride = image(id).w2 \ 4
	group4 = srcW \ 4 - ((srcW AND 3) <> 0)

	'' ---- sprite (slowest) ---- ''
	IF (isSprite) THEN
		OUT gcINDEX, gcReadMapSelect

		lastFullPlane = (srcW AND 3) - 1

		IF (vga.workMemSeg < pg1sg) THEN
			DEF SEG = vga.workMemSeg ' start from target segment
			srcOfs = srcOfs + 16000  ' source is 16000 bytes away
		ELSE
			DEF SEG = pg1sg          ' start from source segment
			dstOfs = dstOfs + 16000  ' destination is 16000 bytes away
		END IF

		FOR p% = 0 TO 3
			OUT scDATA, iSqr(dstPln)
			OUT gcDATA, srcPln
			ofs1 = srcOfs
			ofs2 = dstOfs
			FOR j% = 0 TO srcH - 1
				FOR i% = 0 TO group4 - 1
					pxl = PEEK(ofs1 + i%)
					IF (pxl < 255) THEN POKE (ofs2 + i%), pxl
				NEXT i%
				ofs1 = ofs1 + stride
				ofs2 = ofs2 + 80
			NEXT j%
			dstPln = (dstPln + 1) AND &H3: IF (dstPln = 0) THEN dstOfs = dstOfs + 1
			srcPln = (srcPln + 1) AND &H3: IF (srcPln = 0) THEN srcOfs = srcOfs + 1
			IF (p% = lastFullPlane) THEN
				IF (group4 = 1) THEN EXIT FOR
				group4 = group4 - 1
			END IF
		NEXT p%
		OUT gcINDEX, gcBitMask

	'' ---- solid ---- ''
	ELSE

		'' not plane-aligned (slower) ''
		IF (srcPln + dstPln) THEN
			OUT gcINDEX, gcReadMapSelect
			lastFullPlane = (srcW AND 3) - 1
			' copy group4 pixels, one plane at a time
			FOR p% = 0 TO 3
				OUT gcDATA, srcPln
				OUT scDATA, iSqr(dstPln)
				ofs1 = srcOfs
				ofs2 = dstOfs
				FOR i% = 0 TO srcH - 1
					memCopy pg3sg, ofs1, group4, vga.workMemSeg, ofs2, group4
					ofs1 = ofs1 + stride
					ofs2 = ofs2 + 80
				NEXT i%
				dstPln = (dstPln + 1) AND &H3: IF (dstPln = 0) THEN dstOfs = dstOfs + 1
				srcPln = (srcPln + 1) AND &H3: IF (srcPln = 0) THEN srcOfs = srcOfs + 1
				IF (p% = lastFullPlane) THEN
					IF (group4 = 1) THEN EXIT FOR
					group4 = group4 - 1
				END IF
			NEXT p%
			OUT gcINDEX, gcBitMask

		'' plane-aligned (easy & fast) ''
		ELSE
			' adjust group 4
			group4 = srcW \ 4
			' preserve data in latches
			OUT gcDATA, &H0
			' at least 4 pixels, copy 4 at a time
			IF (group4) THEN
				ofs1 = srcOfs
				ofs2 = dstOfs
				OUT scDATA, &HF
				FOR i% = 0 TO srcH - 1
					memCopy pg3sg, ofs1, group4, vga.workMemSeg, ofs2, group4
					ofs1 = ofs1 + stride
					ofs2 = ofs2 + 80
				NEXT i%
			END IF
			' at most 3 pixels remain, copy all
			IF (srcW AND 3) THEN
				ofs1 = srcOfs + group4
				ofs2 = dstOfs + group4
				OUT scDATA, iSqr(srcW AND 3) - 1
				FOR i% = 0 TO srcH - 1
					memCopy pg3sg, ofs1, 1, vga.workMemSeg, ofs2, 1
					ofs1 = ofs1 + stride
					ofs2 = ofs2 + 80
				NEXT i%
			END IF
			' back to standard write mode
			OUT gcDATA, &HFF
		END IF
	END IF
END SUB

''
'' free image
''
SUB yImgFree (id AS INTEGER)
	DIM start AS INTEGER, ofs AS INTEGER, size AS INTEGER

	SELECT CASE id
	' out of range, bail
	CASE IS < LBOUND(image), IS > UBOUND(image)
		EXIT SUB
	' sprite
	CASE IS < maxBankSize
		DEF SEG = pg1sg
		start = 0
	' solid
	CASE ELSE
		DEF SEG = pg3sg
		start = maxBankSize
	END SELECT
	ofs = image(id).ofs

	' slot already free
	IF (image(id).w = 0) THEN EXIT SUB

	' image size (plane-aligned width)
	size = image(id).size \ 4

	' crush bank offsets
	FOR i% = start TO start + maxBankSize - 1
		IF (image(i%).ofs > ofs) THEN image(i%).ofs = image(i%).ofs - size
	NEXT i%

	' crush charmap offsets or delete linked charmaps
	IF (id < maxBankSize) THEN
		FOR i% = LBOUND(font) TO UBOUND(font)
			IF (font(i%).w) THEN
				IF (font(i%).imgId = id) THEN
					yCMapFree i%
				ELSEIF (font(i%).ofs > ofs) THEN
					font(i%).ofs = font(i%).ofs - size
				END IF
			END IF
		NEXT i%
	END IF

	' crush video memory
	OUT scDATA, &HF
	OUT gcDATA, &H0
	FOR i% = ofs + size TO 15998
		POKE i%, PEEK(i% + 1)
	NEXT i%
	OUT gcDATA, &HFF

	' move memory cursor backward
	IF (id < maxBankSize) THEN ' sprite
		vga.spriteOfs = vga.spriteOfs - size
	ELSE ' solids
		vga.solidOfs = vga.solidOfs - size
	END IF

	' clear slot and kill handle
	image(id).w = 0
	image(id).size = 0
	id = -1
END SUB

''
'' LOAD A MODE 13 BSAVED GRAPHIC FILE
''
FUNCTION yImgFromFile% (filename AS STRING, isSprite AS INTEGER)
	DIM ff AS INTEGER, imgHandle AS INTEGER, dstSeg AS INTEGER, dstOfs AS INTEGER
	DIM signature  AS STRING * 1, junk  AS LONG, imgSize AS INTEGER
	DIM imgWidth AS INTEGER, imgHeight AS INTEGER, imgWidth2 AS INTEGER
	DIM tmpSeg AS INTEGER, scan AS STRING, plnLen AS INTEGER
	DIM scnSeg AS INTEGER, scnOfs AS INTEGER

	' assume failure
	yImgFromFile% = -1

	' open file
	IF (fileExists%(filename) = 0) THEN EXIT FUNCTION
	ff = FREEFILE
	IF (ff = 0) THEN EXIT FUNCTION
	OPEN filename FOR BINARY AS #ff

	' read header
	GET #ff, , signature
	IF (ASC(signature) <> &HFD) THEN CLOSE #ff: EXIT FUNCTION
	GET #ff, , junk ' original memory address of the data (segment:offset)
	GET #ff, , imgSize
	IF ((imgSize < 5) OR (imgSize > (LOF(ff) - 4))) THEN CLOSE #ff: EXIT FUNCTION
	GET #ff, , imgWidth
	IF (imgWidth AND &H7) THEN CLOSE #ff: EXIT FUNCTION
	imgWidth = imgWidth \ 8
	GET #ff, , imgHeight
	' FIXME: I don't now if imgSize is signed or not... but a 320x200 image will
	' fail this test due to INTEGERs being signed.
	IF ((imgWidth * imgHeight) <> (imgSize - 4)) THEN CLOSE #ff: EXIT FUNCTION

	' reserve image
	imgHandle = yImgNew%(imgWidth, imgHeight, isSprite)
	IF (imgHandle = -1) THEN CLOSE #ff: EXIT FUNCTION
	IF (isSprite) THEN
		dstSeg = pg1sg
	ELSE
		dstSeg = pg3sg
	END IF
	dstOfs = image(imgHandle).ofs

	' adjusted length (multiple of 4)
	imgWidth2 = ((imgWidth \ 4) - ((imgWidth AND &H3) <> 0)) * 4

	' reserve 2x the length of a row for conversion buffer
	tmpSeg = memAlloc%(imgWidth2 * 2)
	IF (tmpSeg = 0) THEN CLOSE #ff: EXIT FUNCTION
	DEF SEG = tmpSeg
	scan = SPACE$(imgWidth)
	scnSeg = VARSEG(scan)
	scnOfs = SADD(scan)

	' get offsets to planes in 2nd half of conversion buffer
	plnLen = imgWidth2 \ 4
	imgPl(0) = imgWidth2
	imgPl(1) = imgPl(0) + plnLen
	imgPl(2) = imgPl(1) + plnLen
	imgPl(3) = imgPl(2) + plnLen

	FOR i% = 0 TO imgHeight - 1
		' read scanline
		GET #ff, , scan

		' copy linear data in the 1st half
		memCopy scnSeg, scnOfs, imgWidth, tmpSeg, 0, imgWidth

		' split 1st half into planes in the 2nd half
		tmpOfs = 0
		FOR j% = 0 TO plnLen - 1
			POKE imgPl(0) + j%, PEEK(tmpOfs)
			POKE imgPl(1) + j%, PEEK(tmpOfs + 1)
			POKE imgPl(2) + j%, PEEK(tmpOfs + 2)
			POKE imgPl(3) + j%, PEEK(tmpOfs + 3)
			tmpOfs = tmpOfs + 4
		NEXT j%

		' copy 2nd half to video buffer
		OUT scDATA, &H1: memCopy tmpSeg, imgPl(0), plnLen, dstSeg, dstOfs, plnLen
		OUT scDATA, &H2: memCopy tmpSeg, imgPl(1), plnLen, dstSeg, dstOfs, plnLen
		OUT scDATA, &H4: memCopy tmpSeg, imgPl(2), plnLen, dstSeg, dstOfs, plnLen
		OUT scDATA, &H8: memCopy tmpSeg, imgPl(3), plnLen, dstSeg, dstOfs, plnLen

		' move forward
		dstOfs = dstOfs + plnLen
	NEXT i%
	CLOSE #ff

	' free conversion buffer
	memFree tmpSeg

	' and return handle to newly created image
	yImgFromFile% = imgHandle
END FUNCTION

''
'' GET IMAGE INFO
''
SUB yImgInfo (id AS INTEGER, w AS INTEGER, h AS INTEGER)
	SELECT CASE id
	CASE LBOUND(image) TO UBOUND(image)
		w = image(id).w
		h = image(id).h
	CASE ELSE
		w = 0
		h = 0
	END SELECT
END SUB

''
'' RESERVE IMAGE
''
FUNCTION yImgNew% (w AS INTEGER, h AS INTEGER, isSprite AS INTEGER)
	DIM w2 AS INTEGER, start AS INTEGER, offset AS INTEGER, size AS INTEGER

	' assume we couldn't reserve that image
	yImgNew% = -1

	' invalid image dimensions!
	IF (w < 1) THEN EXIT FUNCTION
	IF (h < 1) THEN EXIT FUNCTION

	' make width plane-aligned
	w2 = (w \ 4 - ((w AND &H3) <> 0)) * 4
	size = w2 * h

	IF (isSprite) THEN
		offset = vga.spriteOfs
		'start = 0
	ELSE
		offset = vga.solidOfs
		start = maxBankSize
	END IF

	' not enough memory!
	IF ((offset + (size \ 4)) > 16000) THEN EXIT FUNCTION

	' find an empty slot
	FOR i% = start TO start + maxBankSize - 1
		IF (image(i%).w < 1) THEN
			' initialize image slot
			image(i%).ofs = offset ' memory location
			image(i%).wr = 0       ' cursor location
			image(i%).w = w        ' width in pixels
			image(i%).h = h        ' height in pixels
			image(i%).w2 = w2      ' width in pixels (upper 4)
			image(i%).size = size  ' buffer size (w2 * h)
			image(i%).x = 0        ' display offset
			image(i%).y = 0        ' display offset
			' move memory cursor forward
			IF (isSprite) THEN
				vga.spriteOfs = vga.spriteOfs + size \ 4
			ELSE
				vga.solidOfs = vga.solidOfs + size \ 4
			END IF
			' return slot index
			yImgNew% = i%
			EXIT FOR
		END IF
	NEXT i%
END FUNCTION

''
'' RETURN PIXEL VALUE
''
FUNCTION yImgPxlGet% (id AS INTEGER)
	DIM wr AS INTEGER

	' assume failture
	yImgPxlGet% = -1

	SELECT CASE id
	' out of range, bail
	CASE IS < LBOUND(image), IS > UBOUND(image)
		EXIT FUNCTION
	' sprite
	CASE IS < maxBankSize
		DEF SEG = pg1sg
	' solid
	CASE ELSE
		DEF SEG = pg3sg
	END SELECT
	wr = image(id).wr
	IF (wr >= image(id).size) THEN EXIT FUNCTION

	' peek pixel
	OUT gcINDEX, gcReadMapSelect
	OUT gcDATA, wr AND 3
	yImgPxlGet% = PEEK(image(id).ofs + wr \ 4)
	OUT gcINDEX, gcBitMask

	' move cursor forward
	wr = wr + 1
	IF ((wr MOD image(id).w2) = image(id).w) THEN
		wr = wr + (image(id).w2 - image(id).w)
	END IF
	image(id).wr = wr
END FUNCTION

''
'' MOVE CURSOR
''
SUB yImgPxlGo (id AS INTEGER, x AS INTEGER, y AS INTEGER)
	' check range
	SELECT CASE id
	CASE IS < LBOUND(image), IS > UBOUND(image)
		EXIT SUB
	END SELECT

	' image undefined
	IF (image(id).w < 1) THEN EXIT SUB

	' stay in horizontal bounds
	SELECT CASE x
	CASE IS >= image(id).w, IS < 0
		EXIT SUB
	END SELECT

	' stay in vertical bounds
	SELECT CASE y
	CASE IS >= image(id).h, IS < 0
		EXIT SUB
	END SELECT

	' alright
	image(id).wr = x + y * image(id).w2
END SUB

''
'' PLOT PIXEL
''
SUB yImgPxlSet (id AS INTEGER, clr AS INTEGER)
	DIM wr AS INTEGER

	SELECT CASE id
	' out of range, bail
	CASE IS < LBOUND(image), IS > UBOUND(image)
		EXIT SUB
	' sprite
	CASE IS < maxBankSize
		DEF SEG = pg1sg
	' solid
	CASE ELSE
		DEF SEG = pg3sg
	END SELECT
	wr = image(id).wr
	IF (wr >= image(id).size) THEN EXIT SUB

	' plot pixel
	OUT scDATA, iSqr(wr AND 3)
	POKE image(id).ofs + wr \ 4, clr

	' move cursor forward
	wr = wr + 1
	IF ((wr MOD image(id).w2) = image(id).w) THEN
		wr = wr + (image(id).w2 - image(id).w)
	END IF
	image(id).wr = wr
END SUB

''
'' PLOT ROW
''
SUB yImgPxlSetRow (id AS INTEGER, srcSeg AS INTEGER, srcOfs AS INTEGER)
	DIM wr AS INTEGER, dstSeg AS INTEGER, dstOfs AS INTEGER, plnLen AS INTEGER
	DIM imgPl(3) AS INTEGER, tmpOfs AS INTEGER, tmpSeg AS INTEGER

	SELECT CASE id
	' out of range, bail
	CASE IS < LBOUND(image), IS > UBOUND(image)
		EXIT SUB
	' sprite
	CASE IS < maxBankSize
		dstSeg = pg1sg
	' solid
	CASE ELSE
		dstSeg = pg3sg
	END SELECT

	' check cursor position
	wr = image(id).wr
	IF (wr >= image(id).size) THEN EXIT SUB

	' move cursor to 1st column of current row
	IF (wr MOD image(id).w2) THEN
		wr = wr - (wr MOD image(id).w2)
	END IF

	' get writing offset
	dstOfs = image(id).ofs + wr \ 4

	' reserve 2x the length of a row
	tmpSeg = memAlloc%(image(id).w2 * 2)
	IF (tmpSeg = 0) THEN EXIT SUB

	' copy linear data in the 1st half
	memCopy srcSeg, srcOfs, image(id).w, tmpSeg, 0, image(id).w

	' get offset to planes in 2nd half
	plnLen = image(id).w2 \ 4
	imgPl(0) = image(id).w2
	imgPl(1) = imgPl(0) + plnLen
	imgPl(2) = imgPl(1) + plnLen
	imgPl(3) = imgPl(2) + plnLen

	' split 1st half into planes in the 2nd half
	DEF SEG = tmpSeg
	tmpOfs = 0
	FOR i% = 0 TO plnLen - 1
		POKE imgPl(0) + i%, PEEK(tmpOfs)
		POKE imgPl(1) + i%, PEEK(tmpOfs + 1)
		POKE imgPl(2) + i%, PEEK(tmpOfs + 2)
		POKE imgPl(3) + i%, PEEK(tmpOfs + 3)
		tmpOfs = tmpOfs + 4
	NEXT i%

	' copy 2nd half to video buffer
	OUT scDATA, &H1: memCopy tmpSeg, imgPl(0), plnLen, dstSeg, dstOfs, plnLen
	OUT scDATA, &H2: memCopy tmpSeg, imgPl(1), plnLen, dstSeg, dstOfs, plnLen
	OUT scDATA, &H4: memCopy tmpSeg, imgPl(2), plnLen, dstSeg, dstOfs, plnLen
	OUT scDATA, &H8: memCopy tmpSeg, imgPl(3), plnLen, dstSeg, dstOfs, plnLen

	' free buffer
	memFree tmpSeg

	' move cursor to next row
	image(id).wr = wr + image(id).w2
END SUB

''
'' SET IMAGE DISPLAY ORIGIN
''
SUB yImgSetOrigin (id AS INTEGER, x AS INTEGER, y AS INTEGER)
	SELECT CASE id
	CASE LBOUND(image) TO UBOUND(image)
		IF (image(id).w > 0) THEN
			image(id).x = x
			image(id).y = y
		END IF
	END SELECT
END SUB

''
'' RETURN THE MEMORY AVAILABLE IN BANKS (OFFSETS ARE PLANE-ALIGNED)
''
SUB yMemInfo (sprBnk AS INTEGER, sldBnk AS INTEGER)
	sprBnk = 16000 - vga.spriteOfs
	sldBnk = 16000 - vga.solidOfs
END SUB

''
'' FREE PALETTE SLOT
''
SUB yPalFree(id as integer)
	SELECT case id
	CASE IS < lbound(clrMap), is > ubound(clrMap)
		exit sub
	case else
		clrMap(id).inUse = 0
		id = -1
	end select
end sub

''
'' RESERVE PALETTE SLOT, RETURNS -1 IF NONE IS AVAILABLE
''
FUNCTION yPalNew%
	DIM ofs as integer

	' assume failure
	yPalNew% = -1

	' find a free spot
	for i% = lbound(clrMap) to ubound(clrMap)
		if (clrMap(i%).inUse = 0) then
			yPalNew% = i%
			clrMap(i%).inUse = -1
			exit for
		end if
	next i%
END FUNCTION

''
'' LOAD PALETTE FROM FILE, RETURNS -1 ON FAIL, A PALETTE INDEX OTHERWISE
''
FUNCTION yPalFromFile%(filename AS STRING)
	dim ff as integer, id as integer, ofs as integer

	' assume failure
	yPalFromFile% = -1

	' get free handle
	id = yPalNew%
	if (id = -1) then exit function

	' open file
	IF (fileExists%(filename) = 0) THEN EXIT FUNCTION
	ff = freefile
	if (ff = 0) THEN EXIT FUNCTION
	open filename for binary as #ff
	if (lof(ff) <> 768) then close #ff: exit function

	' read data
	get #ff,, clrMap(id).clr
	close #ff

	' see if we need to convert to 6 bits
	DEF SEG = varseg(clrMap(id).clr)
	ofs = VARPTR(clrMap(id).clr)
	for i% = 0 to 767
		if (PEEK(ofs + i%) > 63) then 
			fixDepth = -1
			exit for
		end if
	next i%

	' change color depth
	if fixDepth then
		for i% = 0 to 767
			POKE(ofs + i%), PEEK(ofs + i%) \ 4
		next i%
	end if

	' return handle
	yPalFromFile% = id
END FUNCTION

''
'' CONVERT PALETTE TO GRAYSCALE
''
SUB yPalGrayscale(id as integer, first as integer, count as integer)
	dim ofs as integer, g as integer

	' must be an existing palette
	SELECT CASE id
	case is < lbound(clrMap), is > ubound(clrMap)
		exit sub
	end select
	if (clrMap(id).inUse = 0) then exit sub

	' check attribute indices
	select case first
	case is < 0, is > 255
		exit sub
	end select
	select case count + first - 1
	case is < 0, is > 255
		exit sub
	end select

	' convert to gray
	DEF SEG = varseg(clrMap(id).clr)
	ofs = varptr(clrMap(id).clr) + (first * 3)
	for i% = 0 to count -1
		G = (PEEK(ofs) * .299) + (PEEK(ofs + 1) * .587) + (PEEK(ofs + 2) * .114)
		POKE ofs, G
		POKE ofs+1, G
		POKE ofs+2, G
		ofs = ofs + 3
	next i%
END SUB

''
'' SHIFT ATTRIBUTES IN THE CURRENT PALETTE
''
SUB yPalShift(first as integer, last as integer, offset as integer) STATIC
	DIM buffer AS STRING, ofs as integer, count as integer

	select case first
	case is < 0, is > 255
		exit sub
	end select

	select case last
	case is < 0, is > 255
		exit sub
	case is <= first
		exit sub
	end select

	count = last - first + 1
	while (offset < 0)
		offset = offset + count
	wend
	offset = offset MOD count
	buffer = space$(count * 3)

	' copy palette in buffer
	DEF SEG = VARSEG(buffer)
	ofs = SADD(buffer)
	OUT dacREAD, first
	FOR i% = 0 to (count * 3) - 1
		POKE ofs + i%, INP(dacDATA)
	next i%

	' shift
	buffer = right$(buffer, offset * 3) + left$(buffer, len(buffer) - offset * 3)

	' send to DAC
	DEF SEG = VARSEG(buffer)
	ofs = SADD(buffer)
	OUT dacWRITE, first
	FOR i% = 0 TO (count * 3) - 1
		OUT dacDATA, PEEK(ofs + i%)
	NEXT i%
end sub

''
'' STORE CURRENT PALETTE ATTRIBUTES INTO THE SPECIFIED SLOT
''
SUB yPalGet(id as integer, first as integer, count as integer)
	dim ofs as integer

	' must be an existing palette
	SELECT CASE id
	case is < lbound(clrMap), is > ubound(clrMap)
		exit sub
	end select
	if (clrMap(id).inUse = 0) then exit sub

	' check attribute indices
	select case first
	case is < 0, is > 255
		exit sub
	end select
	select case count + first - 1
	case is < 0, is > 255
		exit sub
	end select

	DEF SEG = varseg(clrMap(id).clr)
	ofs = varptr(clrMap(id).clr) + (first * 3)
	OUT dacREAD, first
	FOR i% = 0 TO (count * 3) - 1
		POKE ofs + i%, INP(dacDATA)
	NEXT i%
end sub


''
'' SET PALETTE ATTRIBUTES, THE SOURCE BUFFER MUST BE 3x COUNT
''
SUB yPalApply (id as integer, first AS INTEGER, count AS INTEGER)
	DIM ofs as integer

	' must be an existing palette
	SELECT CASE id
	case is < lbound(clrMap), is > ubound(clrMap)
		exit sub
	end select
	if (clrMap(id).inUse = 0) then exit sub

	' check attribute indices
	select case first
	case is < 0, is > 255
		exit sub
	end select
	select case count + first - 1
	case is < 0, is > 255
		exit sub
	end select

	' send palette to DAC
	DEF SEG = varseg(clrMap(id).clr)
	ofs = varptr(clrMap(id).clr) + (first * 3)
	OUT dacWRITE, first
	FOR i% = 0 TO (count * 3) - 1
		OUT dacDATA, PEEK(ofs + i%)
	NEXT i%
END SUB

''
'' Same as yPrint, except it replaces pixel data by a solid color
''
SUB yPrintColor (id AS INTEGER, clr as integer, x AS INTEGER, y AS INTEGER, msg AS STRING)
	DIM dstOfs AS INTEGER, x1 AS INTEGER, s1 AS INTEGER, s2 AS INTEGER
	DIM mask4 as string * 4, mask8 as string * 8, bitMask as integer

	' keep color in range
	select case clr
	case is < 0, is > 255
		exit sub
	end select

	' stay in range
	SELECT CASE id
	CASE IS < LBOUND(font), IS > UBOUND(font)
		EXIT SUB
	END SELECT
	IF (font(id).w = 0) THEN EXIT SUB

	' outside screen space
	select case y
	case is < 0, is > 24
		exit sub
	end select

	' crop left and right
	select case x
	case is < 0
		s1 = -x1 + 1
		x1 = 0
		s2 = len(msg) - s1 + 1
	case else
		s1 = 1
		x1 = x
		s2 = len(msg)
	end select
	if ((x1 + s2) > (320 \ font(id).w)) then s2 = (320 \ font(id).w) - x1
	if (s2 < 1) then exit sub

	' target video page
	DEF SEG = vga.workMemSeg
	dstOfs = x1 * (font(id).w \ 4) + y * 640

	' SPRITE
	if (font(id).maskSeg) then
		' 4x8 (unrolled loop)
		IF (font(id).w = 4) THEN
			FOR c% = s1 TO s2
				tmp = ASC(MID$(msg, c%, 1))
				SELECT CASE tmp
				CASE font(id).first TO font(id).last
					' get mask for this character
					memCopy font(id).maskSeg, 4 * (tmp - font(id).first),4,varseg(mask4),varptr(mask4),4
					' draw (each byte contains two rows)
					bitMask = asc(mid$(mask4, 1, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs, clr
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 80, clr
					bitMask = asc(mid$(mask4, 2, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+160, clr
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 240, clr
					bitMask = asc(mid$(mask4, 3, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+320, clr
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 400, clr
					bitMask = asc(mid$(mask4, 4, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+480, clr
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 560, clr
				end select
				dstOfs = dstOfs + 1
			next c%
		' 8x8 (unrolled loop)
		else
			FOR c% = s1 TO s2
				tmp = ASC(MID$(msg, c%, 1))
				SELECT CASE tmp
				CASE font(id).first TO font(id).last
					' get mask for this character
					memCopy font(id).maskSeg, 8 * (tmp - font(id).first),8,varseg(mask8),varptr(mask8),8
					' draw (each byte contains one row)
					bitMask = asc(mid$(mask8, 1, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs, clr
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 1, clr
					bitMask = asc(mid$(mask8, 2, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+80, clr
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 81, clr
					bitMask = asc(mid$(mask8, 3, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+160, clr
					OUT scDATA, bitMask \ &h10: POKE dstOfs +161, clr
					bitMask = asc(mid$(mask8, 4, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+240, clr
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 241, clr
					bitMask = asc(mid$(mask8, 5, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+320, clr
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 321, clr
					bitMask = asc(mid$(mask8, 6, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+400, clr
					OUT scDATA, bitMask \ &h10: POKE dstOfs +401, clr
					bitMask = asc(mid$(mask8, 7, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+480, clr
					OUT scDATA, bitMask \ &h10: POKE dstOfs +481, clr
					bitMask = asc(mid$(mask8, 8, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+560, clr
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 561, clr
				end select
				dstOfs = dstOfs + 2
			next c%
		end if

	' SOLID
	else
		' write 4 pixels at a time
		OUT scDATA, &HF
		' 4x8 (unrolled loop)
		IF (font(id).w = 4) THEN
			FOR c% = s1 TO s2
				SELECT CASE ASC(MID$(msg, c%, 1))
				CASE font(id).first TO font(id).last
					POKE dstOfs, clr
					POKE dstOfs + 80, clr
					POKE dstOfs + 160, clr
					POKE dstOfs + 240, clr
					POKE dstOfs + 320, clr
					POKE dstOfs + 400, clr
					POKE dstOfs + 480, clr
					POKE dstOfs + 560, clr
				END SELECT
				dstOfs = dstOfs + 1
			NEXT c%
		' 8x8 (unrolled loop)
		ELSE
			FOR c% = s1 TO s2
				SELECT CASE ASC(MID$(msg, c%, 1))
				CASE font(id).first TO font(id).last
					POKE dstOfs, clr
					POKE dstOfs + 1, clr
					POKE dstOfs + 80, clr
					POKE dstOfs + 81, clr
					POKE dstOfs + 160, clr
					POKE dstOfs + 161, clr
					POKE dstOfs + 240, clr
					POKE dstOfs + 241, clr
					POKE dstOfs + 320, clr
					POKE dstOfs + 321, clr
					POKE dstOfs + 400, clr
					POKE dstOfs + 401, clr
					POKE dstOfs + 480, clr
					POKE dstOfs + 481, clr
					POKE dstOfs + 560, clr
					POKE dstOfs + 561, clr
				END SELECT
				dstOfs = dstOfs + 2
			NEXT c%
		END IF
	end if
END SUB

''
'' print a string on screen using the specified charmap. X is the column, Y is
'' the row. For a standard 8x8 character map, the maximal values would be 39
'' and 24. For a 4x8 character map, 79 and 24.
''
SUB yPrint (id AS INTEGER, x AS INTEGER, y AS INTEGER, msg AS STRING) STATIC
	DIM dstOfs AS INTEGER, srcOfs AS INTEGER, srcOfs2 AS INTEGER
	DIM x1 AS INTEGER, s1 AS INTEGER, s2 AS INTEGER
	DIM mask4 as string * 4, mask8 as string * 8, bitMask as integer

	' stay in range
	SELECT CASE id
	CASE IS < LBOUND(font), IS > UBOUND(font)
		EXIT SUB
	END SELECT
	IF (font(id).w = 0) THEN EXIT SUB

	' outside screen space
	select case y
	case is < 0, is > 24
		exit sub
	end select

	' crop left and right
	select case x
	case is < 0
		s1 = -x1 + 1
		x1 = 0
		s2 = len(msg) - s1 + 1
	case else
		s1 = 1
		x1 = x
		s2 = len(msg)
	end select
	if ((x1 + s2) > (320 \ font(id).w)) then s2 = (320 \ font(id).w) - x1
	if (s2 < 1) then exit sub

	' source & target select
	dstOfs = x1 * (font(id).w \ 4) + y * 640
	srcOfs = image(font(id).imgId).ofs
	IF (vga.workMemSeg < pg1sg) THEN
		DEF SEG = vga.workMemSeg ' start from target segment
		srcOfs = srcOfs + 16000  ' source is 16000 bytes away
	ELSE
		DEF SEG = pg1sg          ' start from source segment
		dstOfs = dstOfs + 16000  ' destination is 16000 bytes away
	END IF

	' preserve data in latches
	OUT gcDATA, &H0

	' SPRITE
	if (font(id).maskSeg) then
		' 4x8 (unrolled loop)
		IF (font(id).w = 4) THEN
			FOR c% = s1 TO s2
				tmp = ASC(MID$(msg, c%, 1))
				SELECT CASE tmp
				CASE font(id).first TO font(id).last
					' pixel offset in video buffer
					srcOfs2 = srcOfs + 8 * (tmp - font(id).first)
					' get mask for this character
					memCopy font(id).maskSeg, 4 * (tmp - font(id).first),4,varseg(mask4),varptr(mask4),4
					' draw (each byte contains two rows)
					bitMask = asc(mid$(mask4, 1, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs, PEEK(srcOfs2)
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 80, PEEK(srcOfs2 + 1)
					bitMask = asc(mid$(mask4, 2, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+160, PEEK(srcOfs2+2)
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 240, PEEK(srcOfs2 + 3)
					bitMask = asc(mid$(mask4, 3, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+320, PEEK(srcOfs2+4)
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 400, PEEK(srcOfs2 + 5)
					bitMask = asc(mid$(mask4, 4, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+480, PEEK(srcOfs2+6)
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 560, PEEK(srcOfs2 + 7)
				end select
				dstOfs = dstOfs + 1
			next c%
		' 8x8 (unrolled loop)
		else
			FOR c% = s1 TO s2
				tmp = ASC(MID$(msg, c%, 1))
				SELECT CASE tmp
				CASE font(id).first TO font(id).last
					' pixel offset in video buffer
					srcOfs2 = srcOfs + 16 * (tmp - font(id).first)
					' get mask for this character
					memCopy font(id).maskSeg, 8 * (tmp - font(id).first),8,varseg(mask8),varptr(mask8),8
					' draw (each byte contains one row)
					bitMask = asc(mid$(mask8, 1, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs, PEEK(srcOfs2)
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 1, PEEK(srcOfs2 + 1)
					bitMask = asc(mid$(mask8, 2, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+80, PEEK(srcOfs2+2)
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 81, PEEK(srcOfs2 + 3)
					bitMask = asc(mid$(mask8, 3, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+160, PEEK(srcOfs2+4)
					OUT scDATA, bitMask \ &h10: POKE dstOfs +161, PEEK(srcOfs2 + 5)
					bitMask = asc(mid$(mask8, 4, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+240, PEEK(srcOfs2+6)
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 241, PEEK(srcOfs2 + 7)
					bitMask = asc(mid$(mask8, 5, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+320, PEEK(srcOfs2+8)
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 321, PEEK(srcOfs2 + 9)
					bitMask = asc(mid$(mask8, 6, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+400, PEEK(srcOfs2+10)
					OUT scDATA, bitMask \ &h10: POKE dstOfs +401, PEEK(srcOfs2 + 11)
					bitMask = asc(mid$(mask8, 7, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+480, PEEK(srcOfs2+12)
					OUT scDATA, bitMask \ &h10: POKE dstOfs +481, PEEK(srcOfs2 + 13)
					bitMask = asc(mid$(mask8, 8, 1))
					OUT scDATA, bitMask AND &hF: POKE dstOfs+560, PEEK(srcOfs2+14)
					OUT scDATA, bitMask \ &h10: POKE dstOfs + 561, PEEK(srcOfs2 + 15)
				end select
				dstOfs = dstOfs + 2
			next c%
		end if

	' SOLID
	ELSE
		' write 4 pixels at a time
		OUT scDATA, &HF
		' 4x8 (unrolled loop)
		IF (font(id).w = 4) THEN
			FOR c% = s1 TO s2
				tmp = ASC(MID$(msg, c%, 1))
				SELECT CASE tmp
				CASE font(id).first TO font(id).last
					srcOfs2 = srcOfs + 8 * (tmp - font(id).first)
					POKE dstOfs, PEEK(srcOfs2)
					POKE dstOfs + 80, PEEK(srcOfs2 + 1)
					POKE dstOfs + 160, PEEK(srcOfs2 + 2)
					POKE dstOfs + 240, PEEK(srcOfs2 + 3)
					POKE dstOfs + 320, PEEK(srcOfs2 + 4)
					POKE dstOfs + 400, PEEK(srcOfs2 + 5)
					POKE dstOfs + 480, PEEK(srcOfs2 + 6)
					POKE dstOfs + 560, PEEK(srcOfs2 + 7)
				END SELECT
				dstOfs = dstOfs + 1
			NEXT c%
		' 8x8 (unrolled loop)
		ELSE
			FOR c% = s1 TO s2
				tmp = ASC(MID$(msg, c%, 1))
				SELECT CASE tmp
				CASE font(id).first TO font(id).last
					srcOfs2 = srcOfs + 16 * (tmp - font(id).first)
					POKE dstOfs, PEEK(srcOfs2)
					POKE dstOfs + 1, PEEK(srcOfs2 + 1)
					POKE dstOfs + 80, PEEK(srcOfs2 + 2)
					POKE dstOfs + 81, PEEK(srcOfs2 + 3)
					POKE dstOfs + 160, PEEK(srcOfs2 + 4)
					POKE dstOfs + 161, PEEK(srcOfs2 + 5)
					POKE dstOfs + 240, PEEK(srcOfs2 + 6)
					POKE dstOfs + 241, PEEK(srcOfs2 + 7)
					POKE dstOfs + 320, PEEK(srcOfs2 + 8)
					POKE dstOfs + 321, PEEK(srcOfs2 + 9)
					POKE dstOfs + 400, PEEK(srcOfs2 + 10)
					POKE dstOfs + 401, PEEK(srcOfs2 + 11)
					POKE dstOfs + 480, PEEK(srcOfs2 + 12)
					POKE dstOfs + 481, PEEK(srcOfs2 + 13)
					POKE dstOfs + 560, PEEK(srcOfs2 + 14)
					POKE dstOfs + 561, PEEK(srcOfs2 + 15)
				END SELECT
				dstOfs = dstOfs + 2
			NEXT c%
		END IF
	END IF
	' back to standard write mode
	OUT gcDATA, &HFF
END SUB

''
'' ENTER MODE Y (320x200, 8 BITS PER PLANE, 4 PLANES)
''
SUB yScreen (enable AS INTEGER)
	DIM regs AS RegTypeX, tmpInt AS INTEGER

	IF (enable) THEN
		regs.ax = &H13                            ' Enter mode 13 (SCREEN 13)
		CALL INTERRUPTX(&H10, regs, regs)

		' Disable chain-4 mode
		OUT scINDEX, scSequencerMemoryMode        ' Sequencer memory mode
		tmpInt = INP(scDATA)                      ' Get current value
		tmpInt = (tmpInt XOR (tmpInt AND 8)) OR 4 ' Bit 3 (Chain-4) off, bit 2 (O/E Dis.) on
		OUT scDATA, tmpInt                        ' <tmpInt> should be &H06

		' Disable long mode
		OUT crtcINDEX2, crtcUnderlineLocation     ' Underline location
		tmpInt = INP(crtcDATA2)                   ' Get current value
		tmpInt = tmpInt XOR (tmpInt AND 64)       ' Turn bit 6 (double-word addressing) off
		OUT crtcDATA2, tmpInt                     ' <tmpInt> should be &H00

		' Enable byte mode
		OUT crtcINDEX2, crtcModeControl           ' Mode control
		tmpInt = INP(crtcDATA2)                   ' Get current value
		tmpInt = tmpInt OR 64                     ' Turn bit 6 (word/byte mode select) on
		OUT crtcDATA2, tmpInt                     ' <tmpInt> should be &HE3

		' These are the expected default values for these registers (so we don't
		' have to change those functions at every turn.) Remember to always return
		' back to this state!
		OUT crtcINDEX2, crtcStartAddressHigh  ' For page flipping
		OUT scINDEX, scMapMask                ' Default write operation
		OUT gcINDEX, gcBitMask                ' Default read operation

		' Flip pages
		vga.viewHiByte = &H0    ' Page 0 starts at A000:0000
		vga.viewMemSeg = pg0sg  ' Page 0 memory segment
		vga.workHiByte = &H7D   ' Page 2 starts at A000:7D00
		vga.workMemSeg = pg2sg  ' Page 2 memory segment

		' Viewport
		yViewport 0, 0, 320, 200

		' For palette operations
		OUT dacPelMask, &HFF    ' DAC - PEL Mask Register

		' Precache square table
		FOR i% = LBOUND(iSqr) TO UBOUND(iSqr)
			iSqr(i%) = 2 ^ i%
		NEXT i%

		' Reinitialize shared variables (for IDE)
		vga.spriteOfs = 0
		vga.solidOfs = 0
		FOR i% = LBOUND(image) TO UBOUND(image)
			image(i%).w = 0
		NEXT i%
		FOR i% = LBOUND(font) TO UBOUND(font)
			font(i%).w = 0
		NEXT i%
	ELSE
		' Return to mode 3 (80x25 text mode)
		regs.ax = &H3
		CALL INTERRUPTX(&H10, regs, regs)
	END IF
END SUB

''
'' CLEAR WORKING PAGE WITH THE SPECIFIED COLOR INDEX
''
SUB yScreenClear (clr AS INTEGER) STATIC
	DIM size AS INTEGER

	OUT scDATA, &HF
	DEF SEG = vga.workMemSeg

	size = 125

	select case clr
	case is < 0, is > 255
		FOR i% = 0 TO size - 1
			POKE i%, 0
		NEXT i%
	case else
		FOR i% = 0 TO size - 1
			POKE i%, clr
		NEXT i%
	end select

	FOR i% = 0 TO 6
		memCopy vga.workMemSeg, 0, size, vga.workMemSeg, size, size
		size = size * 2
	NEXT i%
END SUB

''
'' COPY VISIBLE PAGE TO WORKING PAGE
''
SUB yScreenCopy
	OUT scDATA, &HF ' select all 4 planes (write)
	OUT gcDATA, &H0 ' cancel CPU write (copy latches as-is)
	memCopy vga.viewMemSeg, 0, 16000, vga.workMemSeg, 0, 16000
	OUT gcDATA, &HFF
END SUB

''
'' SET VIEWPORT
''
SUB yViewport (x AS INTEGER, y AS INTEGER, w AS INTEGER, h AS INTEGER)
	SELECT CASE x
	CASE IS < 0
		vga.viewportX1 = 0
	CASE IS > 319
		vga.viewportX1 = 319
	CASE ELSE
		vga.viewportX1 = x
	END SELECT
	vga.viewportX2 = vga.viewportX1 + w - 1

	SELECT CASE y
	CASE IS < 0
		vga.viewportY1 = 0
	CASE IS > 199
		vga.viewportY1 = 199
	CASE ELSE
		vga.viewportY1 = y
	END SELECT
	vga.viewportY2 = vga.viewportY1 + h - 1

	select case vga.viewportX2
	case is < vga.viewportX1
		vga.viewportX2 = vga.viewportX1
	case is > 319
		vga.viewportX2 = 319
	end select

	select case vga.viewportY2
	case is < vga.viewportY1
		vga.viewportY2 = vga.viewportY1
	case is > 199
		vga.viewportY2 = 199
	end select

	vga.viewportW = vga.viewportX2 - vga.viewportX1 + 1
	vga.viewportH = vga.viewportY2 - vga.viewportY1 + 1
END SUB
