DEFINT A-Z

'$INCLUDE: 'QB.BI'
'$INCLUDE: '$YME.BI'

' Run with QB /L $yme

' A rewrite of Sega's 1990 Columns to test a custom graphic library.
' Mike Hawk - 2022

' TODO:
'   Animated highscore table (I desperately need triple buffering for
'   that and it requires a pretty large rewrite of the library. Maybe
'   next time, sorry.)

CONST cellFlash = &H20          ' 32 - flash gem(s)
CONST cellErase = &H40          ' 64 - gem tagged for deletion (explosion frames)
CONST cellVoid = &H80           ' 128 - grid padding (invisible edges)

CONST efEscape = &H1            ' engine flag: escape
CONST efTestGameOver = &H2      ' test game over before player's turn
CONST efPlayThud = &H4          ' play thud sound
CONST efPlaceWait = &H8         ' give a delay before placing the block
CONST efPause = &H10            ' pause - don't update the game
CONST efScreenshot = &H40       ' queued screenshot request
CONST efJoystick = &H80         ' enable joystick support
CONST efFlashDone = &H100       ' flash gem got destroyed
CONST efBlink = &H200           ' blinking characters for score entry

CONST efDwBack = &H1            ' redraw background
CONST efDwGridGems = &H2        ' redraw grid gems
CONST efDwNextTab = &H4         ' redraw next column

CONST efDwReactionTab = &H8     ' redraw reaction score tab
CONST efDwScoreTab = &H10       ' redraw score tab
CONST efDwLevelTab = &H20       ' redraw level tab
CONST efDwJewelsTab = &H40      ' redraw jewel tab

CONST efDwMenu = &H80           ' redraw all menu
CONST efDwMenuItems = &H100     ' redraw items
CONST efDwMenuClear = &H200     ' clear item area
CONST efDwHiScore = &H400       ' redraw hiscore screen
CONST efDwTabs = efDwReactionTab OR efDwJewelsTab OR efDwLevelTab OR efDwScoreTab OR efDwNextTab

CONST mfRedraw = &H1            ' menu item - redraw
CONST mfChangeMenu = &H2        ' menu item jumps to another menu
CONST mfExitMenu = &H4          ' menu item starts a new game or quits to DOS
CONST mfCounter = &H8           ' menu item can be modified
CONST mfScore = &H10            ' menu item gives best time for flash mode
CONST mfClearScores = &H20      ' reset scores and jump to the specified menu

CONST sndEffect = 0             ' sound effects only (for eng.sndMode)
CONST sndMusic = 1              ' music only (for eng.sndMode)
CONST sndNone = 2               ' no sound or music (for eng.sndMode)

CONST sndThud = 0               ' sound indices
CONST sndExplo = 1
CONST sndPlace = 2
CONST sndShift = 3
CONST sndLevel = 4
CONST sndExplo2 = 5
CONST sndExplo3 = 6

CONST trPlayer = 0              ' player turn
CONST trExplosion = 1           ' computer turn: remove gems
CONST trStack = 2               ' computer turn: stack gems
CONST trGameOver = 3            ' game over
CONST trStart = 4               ' starting fanfar
CONST trWellDone = 5            ' flash columns
CONST trInputName = 6           ' input name

CONST blkGEMS = 0               ' block type (see blocks() array)
CONST blkDICE = 1               ' block type (see blocks() array)
CONST blkCARDS = 2              ' block type (see blocks() array)
CONST blkSQUARES = 3            ' block type (see blocks() array)
CONST blkFRUITS = 4             ' block type (see blocks() array)

CONST minBlockId = 1            ' blocks' lowest index
CONST maxBlockId = 6            ' blocks' highest index
CONST mgcBlockId = 7            ' magic blocks index

CONST kbDisable = 0             ' Go back to initial ISR (retain memory)
CONST kbEnable = 1              ' Install new ISR, reserve memory if needed
CONST kbUninstall = 2           ' Free memory for ISR, restore initial ISR

CONST msLoop = &H1              ' special music flag

CONST cLeft = &H1               ' user input flags
CONST cRight = &H2
CONST cUp = &H4
CONST cDown = &H8
CONST cAction = &H10
CONST cPause = &H20
CONST cSound = &H40
CONST cEscape = &H80
CONST cScreenshot = &H100

CONST imgBlock1 = 0             ' must appear before imgBlock2
CONST imgBlock2 = 1             ' must appear after imgBlock1
CONST imgCell = 2
CONST imgCellRow = 8
CONST imgGameOver = 9
CONST imgTabNext = 3
CONST imgTabReaction = 4
CONST imgTabScore = 5
CONST imgTabLevel = 6
CONST imgTabJewels = 7
CONST imgJems = 10              ' this sprawls 36 indices!
CONST imgWellDone = 95
CONST imgTabSkill = 111
CONST imgTabBest = 112
CONST imgGemGlow = 120          ' this sprawls 7 indices!

CONST iColor = 15           ' "interactive color" (menus) either 241 or 15... IDK

TYPE rankType1              ' OFS SZE  DESCRIPTION
	pl AS STRING * 4          ' 000 004  player name (3 chars displayed)
	score AS LONG             ' 004 004  score
	jewels AS INTEGER         ' 008 002  jewels destroyed
	level AS INTEGER          ' 010 002  level achieved
END TYPE                    ' 12 BYTES - HISCORE TABLE

TYPE rankType2              ' OFS SZE  DESCRIPTION
	pl AS STRING * 4          ' 000 004  player name (3 chars displayed)
	tmr AS INTEGER            ' 004 002  best time (in seconds)
END TYPE                    ' 6 BYTES - BEST TIME (FLASH COLUMNS)

TYPE itemType               ' OFS SZE  DESCRIPTION
	msg AS STRING * 28        ' 000 028  Display string
	flags AS INTEGER          ' 028 002  Special flags
	y AS INTEGER              ' 030 002  Display row
END TYPE                    ' 32 BYTES - MENU ITEMS

TYPE engineType             ' OFS SZE  DESCRIPTION
	blockType AS INTEGER      ' 000 002  block type (0 to 4)
	keyCtrl AS INTEGER        ' 002 002  previous key
	keyCD AS INTEGER          ' 004 002  key countdown
	keyDelay AS INTEGER       ' 006 002  reset value for key countdown
	keySegm AS INTEGER        ' 008 002  memory location to keyboard presses
	track AS INTEGER          ' 010 002  currently playing track
	plTrack AS INTEGER        ' 012 002  user music
	flags AS INTEGER          ' 014 002  generic signals and reminders
	dFlags AS INTEGER         ' 016 002  rendering/drawing signals and reminders
	pushDelay AS INTEGER      ' 018 002  push countdown reset value
	waitDelay AS INTEGER      ' 020 002  wait before placing block for good
	stackDelay AS INTEGER     ' 022 002  stacking countdown reset value
	exploDelay AS INTEGER     ' 024 002  explosion countdown reset value
	countdown AS INTEGER      ' 026 002  general countdown
	frames AS INTEGER         ' 028 002  explosion frame
	x AS INTEGER              ' 030 002  column x position (whole grid, 16 pixels)
	y AS INTEGER              ' 032 002  column y position (half grid, 7 pixels)
	sndMode AS INTEGER        ' 034 002  sound mode (fx, music, none)
	gamMode AS INTEGER        ' 036 002  game mode (0: original, 1: flash)
	turn AS INTEGER           ' 038 002  gameplay turn (player, explosion, stack)
	chainReact AS INTEGER     ' 040 002  chain reaction counter
	scoreTmp AS LONG          ' 042 004  temporary score (current reaction)
	score AS LONG             ' 046 004  total score
	level AS INTEGER          ' 050 002  current level
	levelNext AS INTEGER      ' 052 002  number of gems to next level
	levelPts AS INTEGER       ' 054 002  points per gem eliminated on this level
	gems AS INTEGER           ' 056 002  number of destroyed gems
	skill AS INTEGER          ' 058 002  skill level (0: NOVICE, 1: AMATEUR, 2: PRO)
	levelStart AS INTEGER     ' 060 002  starting level (0: EASY, 5: MEDIUM, 10: HARD)
	floor AS INTEGER          ' 062 002  floor height (junk)
	sListSeg AS INTEGER       ' 064 002  points to search list (indices of aligned cells)
	sListLen AS INTEGER       ' 066 002  length of the search list, in bytes, including grid copy
	tFlashCD AS INTEGER       ' 068 002  flash columns timer countdown
	tFlash AS INTEGER         ' 070 002  flash columns timer (in seconds)
	glowFrame AS INTEGER      ' 072 002  gem flash frame
	glowCD AS INTEGER         ' 074 002  gem flash timer (increases)
	plName AS STRING * 4      ' 076 004  player name (for highscore)
	plNameOfs AS INTEGER      ' 080 002  offset in player name
	menuIndex AS INTEGER
	grid AS STRING * 256      ' 082 256  128 cells + mirror, edges are hidden (2 up, 1 down, 1 left, 1 right)
END TYPE                    ' 338 BYTES - Generic game/engine stuff

DECLARE SUB initProg ()
DECLARE SUB initScores ()

DECLARE SUB gameLoop ()
DECLARE FUNCTION menuLoop% ()
DECLARE SUB hiscoreLoop ()

DECLARE SUB gameRender ()

DECLARE SUB columnGetNext ()
DECLARE FUNCTION columnMove% (x AS INTEGER, y AS INTEGER)

DECLARE SUB gridClearAll ()
DECLARE SUB gridAdd ()
DECLARE SUB gridTest1 ()
DECLARE SUB gridTest2 ()
DECLARE SUB gridPrune ()
DECLARE FUNCTION gridStack% ()
DECLARE FUNCTION gridJunkRow$ (numPieces AS INTEGER)

DECLARE SUB screenshot ()
DECLARE SUB playSound (id AS INTEGER)
DECLARE SUB playMusic (id AS INTEGER)
DECLARE FUNCTION loadTrack$ (filename AS STRING)

DECLARE FUNCTION getPIT& ()
DECLARE FUNCTION keyboardInit% (mode AS INTEGER)
DECLARE SUB commonProcess ()

DIM SHARED img(127) AS INTEGER      ' images handles
DIM SHARED cMap(7) AS INTEGER       ' fonts handles
DIM SHARED blocks(6) AS INTEGER     ' binds block IDs to indices in img()
DIM SHARED thisCol(2) AS INTEGER    ' current player column
DIM SHARED nextCol(2) AS INTEGER    ' next column
DIM SHARED music(7) AS STRING       ' PLAY string buffers
DIM SHARED musicFlag(7) AS INTEGER  ' special music attributes
DIM SHARED rank1(8) AS rankType1    ' high score table (original)
DIM SHARED rank2(239) AS rankType2  ' best time (flash)
DIM SHARED eng AS engineType        ' engine "namespace"

'' INITIALIZE ''
initProg                                  ' draw splash screen, load graphics
eng.keySegm = keyboardInit%(kbEnable)     ' initialize custom keyboard ISR

'' MENU/GAME LOOP ''
DO
	i% = menuLoop%
	SELECT CASE i%
	CASE 0
		EXIT DO
	CASE 1
		hiscoreLoop
	CASE ELSE
		eng.gamMode = i% - 2
		gameLoop
		IF (eng.gamMode = 0) THEN hiscoreLoop ' only in original mode
	END SELECT
LOOP

'' SAVE SCORES ''
OPEN "FLASH.HI" FOR BINARY AS #1
FOR i% = 0 TO 239
	PUT #1, , rank2(i%)
NEXT i%
CLOSE #1
OPEN "SCORES.HI" FOR BINARY AS #1
FOR i% = 0 TO 8
	PUT #1, , rank1(i%)
NEXT i%
CLOSE #1

'' SHUT DOWN ''
eng.keySegm = keyboardInit%(kbUninstall)  ' restore ISR
memFree eng.sListSeg                      ' free search list
FOR i% = LBOUND(cMap) TO UBOUND(cMap)     ' free all fonts
	yCMapFree cMap(i%)
NEXT i%
yScreen 0                                 ' return to text mode
END                                       ' and we're done

''
'' Give next column to player, pick a new random column
''
SUB columnGetNext
	' give next column to player
	thisCol(0) = nextCol(0)
	thisCol(1) = nextCol(1)
	thisCol(2) = nextCol(2)

	' reset column position
	eng.x = 3
	eng.y = 4

	' generate another column
	IF ((eng.levelStart > 4) AND (INT(RND * 512) = 123) AND (eng.gamMode = 0)) THEN
		' magic gem
		nextCol(0) = mgcBlockId
		nextCol(1) = mgcBlockId
		nextCol(2) = mgcBlockId
	ELSE
		' create new column (NOVICE, AMATEUR, PRO)
		nextCol(0) = INT(RND * (4 + eng.skill)) + minBlockId
		nextCol(1) = INT(RND * (4 + eng.skill)) + minBlockId
		nextCol(2) = INT(RND * (4 + eng.skill)) + minBlockId
	END IF
END SUB

''
'' Due to the way the game works, we only need to test the base (lowest tile)
'' of the column before a move (since all gems are stacked at the bottom of
'' the grid, it's impossible to have blank spaces in-between gems,) this
'' function returns 0 on collision.
''
FUNCTION columnMove% (x AS INTEGER, y AS INTEGER) STATIC
	DIM doMove AS INTEGER

	IF (x) THEN ' horizontal test
		doMove = (ASC(MID$(eng.grid, 1 + ((eng.y + (eng.y AND 1)) \ 2) * 8 + (eng.x + x), 1)) = 0)
	ELSEIF (y) THEN ' vertical test
		doMove = (ASC(MID$(eng.grid, 1 + (y + (eng.y - (eng.y AND 1)) \ 2) * 8 + eng.x, 1)) = 0)
	END IF

	' move?
	IF (doMove) THEN
		eng.x = eng.x + x
		eng.y = eng.y + y
	END IF

	' return true if the column moved
	columnMove% = doMove
END FUNCTION

''
'' Invoked early in timed loops (music & controls)
''
SUB commonProcess STATIC
	DIM ctrl AS INTEGER, playTrack AS INTEGER
	DIM playPos1 AS INTEGER, playPos2 AS INTEGER

	'' MUSIC PLAYBACK ''
	IF (eng.track <> playTrack) THEN
		playPos1 = 0
		playTrack = eng.track
	END IF
	IF (eng.sndMode <> sndMusic) THEN
		playPos1 = 0
		GOTO musicSkip
	END IF
	IF (PLAY(0) < 3) THEN
		IF (playPos1 = LEN(music(eng.track))) THEN
			IF (musicFlag(eng.track) AND msLoop) THEN
				playPos1 = 0
			ELSE
				GOTO musicSkip
			END IF
		END IF
		playPos2 = INSTR(playPos1 + 1, music(eng.track), " ")
		PLAY MID$(music(eng.track), playPos1 + 1, playPos2 - playPos1)
		playPos1 = playPos2
	END IF
musicSkip:

	'' CONTROLS (JOYSTICK) ''
	ctrl = 0
	IF (eng.flags AND efJoystick) THEN
		' left or right
		IF ((STICK(0) - 128) < -64) THEN
			ctrl = ctrl OR cLeft
		ELSEIF ((STICK(0) - 128) > 64) THEN
			ctrl = ctrl OR cRight
		END IF
		' up or down
		IF ((STICK(1) - 128) < -64) THEN
			ctrl = ctrl OR cUp
		ELSEIF ((STICK(1) - 128) > 64) THEN
			ctrl = ctrl OR cDown
		END IF
		' action
		IF (STRIG(1) OR STRIG(5) OR STRIG(3) OR STRIG(7)) THEN
			ctrl = ctrl OR cAction
		END IF
	END IF

	'' CONTROLS (KEYBOARD) ''
	DEF SEG = eng.keySegm
	' left or right
	IF PEEK(&H4B) THEN
		ctrl = ctrl OR cLeft
	ELSEIF PEEK(&H4D) THEN
		ctrl = ctrl OR cRight
	END IF
	' up or down
	IF PEEK(&H48) THEN
		ctrl = ctrl OR cUp
	ELSEIF PEEK(&H50) THEN
		ctrl = ctrl OR cDown
	END IF
	' action
	IF PEEK(&H1C) THEN
		ctrl = ctrl OR cAction
	END IF

	'' COMMON CONTROLS ''
	IF PEEK(&H58) THEN ' F12
		ctrl = ctrl OR cScreenshot
		IF ((eng.keyCtrl AND cScreenshot) = 0) THEN eng.flags = eng.flags OR efScreenshot
	ELSEIF PEEK(&H1) THEN ' escape
		ctrl = ctrl OR cEscape
		IF ((eng.keyCtrl AND cEscape) = 0) THEN eng.flags = eng.flags OR efEscape
	ELSEIF PEEK(&H19) THEN ' P
		ctrl = ctrl OR cPause
	ELSEIF PEEK(&H1F) THEN ' S
		ctrl = ctrl OR cSound
		IF ((eng.keyCtrl AND cSound) = 0) THEN
			playPos1 = 0
			eng.sndMode = (eng.sndMode + 1) MOD 3
		END IF
	END IF

	'' LONG-PRESS DELAY ''
	' NOTE: pushing a block down sets eng.keyCD to 1 rather than 0 so the next
	' evaluation doesn't reset it to eng.keyDelay (and instead sets it to 0.)
	IF (eng.keyCtrl = ctrl) THEN
		IF (eng.keyCD = 0) THEN eng.keyCD = eng.keyDelay
		eng.keyCD = eng.keyCD + (eng.keyCD > 0)
	ELSE
		eng.keyCD = 0
	END IF
	eng.keyCtrl = ctrl
END SUB

''
'' GAME LOOP (FOR BOTH ORIGINAL AND FLASH MODES)
''
SUB gameLoop
	DIM tmp AS INTEGER, ofsX AS INTEGER, ofsY AS INTEGER
	DIM rankIndex AS INTEGER, rankSave AS INTEGER
	DIM tPrevious AS LONG, tLag AS LONG, tCurrent AS LONG, tElapsed AS LONG

	' blocks lookup table
	FOR i% = 0 TO 5
		blocks(i%) = img(imgJems + (eng.blockType * 6) + i%) ' standard
	NEXT i%
	blocks(6) = img(40) ' magic

	eng.keyDelay = 10   ' ticks between identical key presses (except down)

	eng.pushDelay = 60  ' number of game ticks before pushing column down
	eng.waitDelay = 30  ' number of ticks before the column snaps in place
	eng.stackDelay = 5  ' how long before blocks fall after a reaction
	eng.exploDelay = 3  ' ticks between two frames of animation (explosions)
	eng.frames = 0      ' starting frame for explosions
	eng.levelNext = 50  ' destroyed gems before next level (countdown)

	eng.score = 0       ' reset score (might change according to level
	eng.scoreTmp = 0    ' reaction score (resets after chain reactions)
	eng.chainReact = 0  ' chain reaction count (flash mode may not reset that)
	eng.gems = 0        ' number of gems removed

	eng.level = eng.levelStart
	eng.levelPts = (eng.level + 1) * 10
	eng.pushDelay = eng.pushDelay - (eng.level * 10)
	IF (eng.pushDelay < 4) THEN eng.pushDelay = 4

	' default player name
	eng.plName = "AAA"
	eng.plNameOfs = 1

	' reset timer for flash mode
	eng.tFlashCD = 0
	eng.tFlash = 0
	eng.glowCD = 0
	eng.glowFrame = 0

	' set turn to start fanfar
	eng.countdown = 240
	eng.turn = trStart

	' initialize starting fanfar playback
	eng.track = 1
 
	' initialize a new grid (includes junk floor)
	gridClearAll

	' get next column (twice, because the next column is initially blank)
	columnGetNext

	' clear all flags, except...
	eng.flags = eng.flags AND efJoystick

	' force draw everything
	eng.dFlags = efDwBack OR efDwTabs OR efDwGridGems

	tPrevious = getPIT&
	DO
		' update timer
		tCurrent = getPIT&
		tElapsed = tCurrent - tPrevious
		tPrevious = tCurrent
		tLag = tLag + tElapsed

		' update game logic for 78 clock ticks for every
		' slice of 78 ticks we skipped during rendering.
		WHILE (tLag > 78)

			' decrease countdown to whatever mode we're in
			eng.countdown = eng.countdown + (eng.countdown > 0)

			' capture keys
			commonProcess

			' process always-on keys
			IF (eng.keyCD = 0) THEN
				IF (eng.keyCtrl AND cPause) THEN
					SELECT CASE eng.turn
					CASE trPlayer, trExplosion, trStack
						IF (eng.flags AND efPause) THEN
							MID$(eng.grid, 129, 128) = STRING$(128, 255)
							eng.dFlags = eng.dFlags OR efDwGridGems
						END IF
						eng.flags = eng.flags XOR efPause
					END SELECT
				END IF
			END IF

			IF (eng.flags AND efPause) THEN
				' do nothing
			ELSE
				' timer for flash mode
				IF (eng.gamMode = 1) THEN
					SELECT CASE eng.turn
					CASE trPlayer, trExplosion, trStack
						IF (eng.tFlashCD = 59) THEN
							eng.tFlash = eng.tFlash + 1
							eng.dFlags = eng.dFlags OR efDwReactionTab
							eng.tFlashCD = 0
						ELSE
							eng.tFlashCD = eng.tFlashCD + 1
						END IF
					END SELECT
					IF (eng.glowCD) THEN
						eng.glowCD = eng.glowCD - 1
					ELSE
						eng.glowFrame = (eng.glowFrame + 1) AND &HF
						eng.glowCD = 8
						IF (eng.glowFrame < 8) THEN eng.dFlags = eng.dFlags OR efDwGridGems
					END IF
				END IF

				SELECT CASE eng.turn
				CASE trStart
					IF (eng.countdown = 0) THEN
						' set music to player track
						eng.track = eng.plTrack
						' get new column
						columnGetNext
						' set turn to player
						eng.countdown = eng.pushDelay
						eng.turn = trPlayer
						' force refresh grid
						MID$(eng.grid, 129, 128) = STRING$(128, 255)
						eng.dFlags = eng.dFlags OR efDwGridGems OR efDwNextTab
					END IF

				CASE trGameOver, trWellDone
					IF (eng.countdown = 0) THEN
						eng.countdown = eng.exploDelay
						eng.frames = eng.frames + 1
						SELECT CASE eng.frames
						CASE IS < 32
							IF (eng.frames AND 1) THEN playSound sndExplo3
						CASE 36
							eng.frames = 32
						END SELECT
						IF ((eng.frames AND &H3) = 3) THEN yPalShift 247, 254, 2
						
						' TODO: CHECK IF WE HAVE A REASON TO INPUT NAME
						IF (eng.keyCtrl AND cAction) THEN

							'' TEST SCORE - FLASH COLUMNS ''
							IF (eng.gamMode) THEN
								' only if we achieved to remove all flashing gems!
								IF (eng.turn = trWellDone) THEN
									' get rank index
									rankIndex = eng.skill + ((eng.floor - 2) * 3) + (eng.level * 24)
									' better time?
									rankSave = ((eng.tFlash <= rank2(rankIndex).tmr) OR (rank2(rankIndex).tmr = 0))
								END IF

							'' TEST SCORE - ORIGINAL MODE ''
							ELSE
								' get rank index
								FOR rankIndex = 0 TO 8
									' better score?
									IF (eng.score >= rank1(rankIndex).score) THEN
										rankSave = -1
										EXIT FOR
									END IF
								NEXT rankIndex
							END IF

							' input name if we got a score, leave otherwise
							IF (rankSave) THEN
								' force refresh grid (clear all)
								MID$(eng.grid, 1, 128) = STRING$(128, 0)
								MID$(eng.grid, 129, 128) = STRING$(128, 255)
								eng.dFlags = eng.dFlags OR efDwGridGems
								' input name
								eng.turn = trInputName
							ELSE
								eng.flags = eng.flags OR efEscape
							END IF
						END IF
					END IF

				CASE trInputName
					IF (eng.countdown = 0) THEN
						eng.countdown = 20 ' blinking delay
						eng.flags = eng.flags XOR efBlink
					END IF
					IF (eng.keyCD = 0) THEN
						IF (eng.keyCtrl AND cLeft) THEN
							eng.plNameOfs = eng.plNameOfs + (eng.plNameOfs > 1)
						ELSEIF (eng.keyCtrl AND cRight) THEN
							eng.plNameOfs = eng.plNameOfs - (eng.plNameOfs < 3)
						END IF
						IF (eng.keyCtrl AND cUp) THEN
							tmp = ASC(MID$(eng.plName, eng.plNameOfs, 1))
							MID$(eng.plName, eng.plNameOfs, 1) = CHR$(tmp - (tmp < 90))
						ELSEIF (eng.keyCtrl AND cDown) THEN
							tmp = ASC(MID$(eng.plName, eng.plNameOfs, 1))
							MID$(eng.plName, eng.plNameOfs, 1) = CHR$(tmp + (tmp > 32))
						END IF
						' save score
						IF (eng.keyCtrl AND cAction) THEN
							'' FLASH COLUMNS ''
							IF (eng.gamMode) THEN
								' replace best time
								rank2(rankIndex).tmr = eng.tFlash
								rank2(rankIndex).pl = eng.plName
							'' ORIGINAL MODE ''
							ELSE
								' push score down the list
								FOR i% = 8 TO rankIndex + 1 STEP -1
									rank1(i%) = rank1(i% - 1)
								NEXT i%
								' insert new score
								rank1(rankIndex).pl = eng.plName
								rank1(rankIndex).score = eng.score
								rank1(rankIndex).jewels = eng.gems
								rank1(rankIndex).level = eng.level
							END IF
							eng.flags = eng.flags OR efEscape
						END IF
					END IF

				CASE trPlayer
					' before anything else, make sure the player is allowed to play his
					' turn: if the 2nd row contains any block or the starting point is
					' already occupied, set the game over flag.
					IF (eng.flags AND efTestGameOver) THEN
						' cancel check for the tick
						eng.flags = eng.flags AND (NOT efTestGameOver)
						' display next block, clear "reaction score" tab
						eng.dFlags = eng.dFlags OR (efDwNextTab OR efDwReactionTab)
						' reset temporary score (chain reactions are over by now)
						eng.scoreTmp = 0
						eng.chainReact = 0
						' allow some time before taking DOWN input again
						eng.keyCD = eng.keyDelay * 4
						' test the row right above the view port, test the entry point
						IF ((MID$(eng.grid, 10, 6) <> STRING$(6, 0)) OR (MID$(eng.grid, 1 + eng.x + (eng.y \ 2) * 8, 1) <> CHR$(0))) THEN
							' force refresh gems
							MID$(eng.grid, 129, 128) = STRING$(128, 255)
							eng.dFlags = eng.dFlags OR efDwGridGems
							' initalize game over flame wall (resetting the countdown now prevents game inputs)
							eng.frames = 0
							eng.countdown = eng.exploDelay
							eng.turn = trGameOver
							' game over jingle playback
							eng.track = 2
						ELSE
							eng.countdown = eng.pushDelay
						END IF
					END IF

					'' process player input
					IF (eng.keyCD = 0) THEN
						' shift column
						IF (eng.keyCtrl AND cAction) THEN
							tmp = thisCol(2)
							thisCol(2) = thisCol(1)
							thisCol(1) = thisCol(0)
							thisCol(0) = tmp
							playSound sndShift

						' rush down
						ELSEIF (eng.keyCtrl AND cDown) THEN
							IF ((eng.flags AND efPlaceWait) = 0) THEN
								IF (eng.gamMode = 0) THEN
									eng.score = eng.score + 1
									eng.dFlags = eng.dFlags OR efDwScoreTab
								END IF
							END IF
							eng.countdown = 0 ' force block down
							eng.keyCD = 1 ' if keypress = oldkeypress and keyCD is not 0...
						END IF

						' move left
						IF (eng.keyCtrl AND cLeft) THEN
							tmp = columnMove%(-1, 0)

						' move right
						ELSEIF (eng.keyCtrl AND cRight) THEN
							tmp = columnMove%(1, 0)
						END IF
					END IF

					'' move block down automatically
					IF (eng.countdown = 0) THEN
						' no collision, cancel "efPlaceWait", reset push countdown
						IF (columnMove%(0, 1)) THEN
							eng.flags = eng.flags AND (NOT efPlaceWait)
							eng.countdown = eng.pushDelay

						' collision,...
						ELSE
							' player already waited, place now.
							IF (eng.flags AND efPlaceWait) THEN
								eng.flags = eng.flags XOR efPlaceWait
								IF (thisCol(2) = mgcBlockId) THEN
									gridAdd
									gridTest2
									columnGetNext
									eng.dFlags = eng.dFlags OR efDwGridGems
									eng.flags = eng.flags OR efTestGameOver
								ELSE
									gridAdd
									columnGetNext
									gridTest1
									eng.dFlags = eng.dFlags OR efDwGridGems
									eng.flags = eng.flags OR efTestGameOver
									playSound sndPlace
								END IF
							' allow a short delay to move/shift the column
							ELSE
								eng.flags = eng.flags OR efPlaceWait
								eng.countdown = eng.waitDelay
							END IF
						END IF
					END IF

				'' explosion frames for gems tagged for removal
				CASE trExplosion
					IF (eng.countdown = 0) THEN
						eng.countdown = eng.exploDelay
						eng.frames = eng.frames + 1
						eng.dFlags = eng.dFlags OR efDwGridGems
						IF (eng.frames = 16) THEN
							eng.frames = 0
							gridPrune
							' set turn to stack
							eng.countdown = eng.stackDelay
							eng.turn = trStack
							' add score to total or end game (flash columns)
							IF (eng.gamMode = 0) THEN
								eng.score = eng.score + eng.scoreTmp
								eng.dFlags = eng.dFlags OR efDwScoreTab
							ELSEIF (eng.flags AND efFlashDone) THEN
								eng.frames = 0
								eng.countdown = eng.exploDelay
								eng.turn = trWellDone
							END IF
						END IF
					END IF

				'' stack gems down
				CASE trStack
					IF (eng.countdown = 0) THEN
						IF (gridStack%) THEN
							eng.countdown = eng.stackDelay
							eng.flags = eng.flags OR efPlayThud
						ELSE
							IF (eng.flags AND efPlayThud) THEN
								eng.flags = eng.flags AND (NOT efPlayThud)
								playSound sndThud
							END IF
							' set turn to player
							eng.countdown = eng.pushDelay
							eng.turn = trPlayer
							' test grid, if we got a hit, player turn is cancelled
							gridTest1
						END IF
						eng.dFlags = eng.dFlags OR efDwGridGems
					END IF
				END SELECT
			END IF

			' A game tick is 78 PIT ticks, 4660 PIT ticks = 1 second
			tLag = tLag - 78
		WEND

		'' RENDERING, display next frame ''
		gameRender
		yFlip
		IF (eng.flags AND efScreenshot) THEN
			screenshot
			eng.flags = eng.flags AND (NOT efScreenshot)
		END IF

		' ready back buffer
		yScreenCopy
		IF (eng.turn = trPlayer) THEN
			ofsX = (eng.x - 1) * 16
			ofsY = (eng.y + (eng.y AND 1) - 4) * 7
			yImgDraw img(imgCell), ofsX, ofsY
			yImgDraw img(imgCell), ofsX, ofsY - 14
			yImgDraw img(imgCell), ofsX, ofsY - 28
			IF (eng.y AND 1) THEN yImgDraw img(imgCell), ofsX, ofsY - 42
		END IF

		' preserve grid state of previous render
		MID$(eng.grid, 129, 128) = LEFT$(eng.grid, 128)

		' cancel all redraw flags
		eng.dFlags = 0

	LOOP UNTIL (eng.flags AND efEscape)
	eng.flags = eng.flags AND (NOT efEscape)
END SUB

''
'' Render gameplay frame
''
SUB gameRender STATIC
	DIM ofsX AS INTEGER, ofsY AS INTEGER, ofsS AS INTEGER
	DIM tmp AS INTEGER, tmpStr AS STRING, tmpStr2 AS STRING

	'' redraw backdrop
	IF (eng.dFlags AND efDwBack) THEN
		' reset viewport
		yViewport 0, 0, 320, 200

		' big tiles
		FOR y% = 0 TO 7
			ofsY = y% * 27
			ofsX = 0
			FOR x% = 0 TO 4
				yImgDraw img(imgBlock1 + (y% AND 1)), ofsX, ofsY
				yImgDraw img(imgBlock1 + (1 XOR (y% AND 1))), ofsX + 32, ofsY
				ofsX = ofsX + 64
			NEXT x%
		NEXT y%

		' grid edges (top-bottom)
		yImgDraw img(53), 110, 7
		yImgDraw img(54), 178, 7
		yImgDraw img(54), 128, 7
		yImgDraw img(53), 160, 7
		yImgDraw img(54), 110, 191
		yImgDraw img(53), 178, 191
		yImgDraw img(53), 128, 191
		yImgDraw img(54), 160, 191

		' grid edges (left-right)
		yImgDraw img(51), 110, 7
		yImgDraw img(52), 206, 7
		yImgDraw img(52), 110, 164
		yImgDraw img(51), 206, 164
		FOR y% = 1 TO 6
			ofsY% = y% * 27
			yImgDraw img(51 + (y% AND 1)), 110, ofsY
			yImgDraw img(51 + (1 XOR (y% AND 1))), 206, ofsY
		NEXT y%

		'' FLASH COLUMNS ''
		IF (eng.gamMode) THEN
			yImgDraw img(imgTabReaction), 61, 71  ' timer tab
			yImgDraw img(imgTabBest), 37, 94      ' best tab
			yImgDraw img(imgTabLevel), 37, 126    ' level tab
			yImgDraw img(imgTabSkill), 37, 158    ' skill tab

		'' ORIGINAL MODE ''
		ELSE
			yImgDraw img(imgTabReaction), 61, 71  ' reaction tab
			yImgDraw img(imgTabScore), 37, 94     ' score tab
			yImgDraw img(imgTabLevel), 37, 126    ' level tab
			yImgDraw img(imgTabJewels), 37, 158   ' jewel tab
		END IF

		' reset viewport to grid, draw grid
		yViewport 112, 9, 96, 182
		FOR y% = 0 TO 12
			yImgDraw img(imgCellRow), 0, y% * 14
		NEXT y%
	END IF

	IF (eng.dFlags AND efDwTabs) THEN
		' common to both modes
		IF (eng.dFlags AND efDwNextTab) THEN
			' set viewport
			yViewport 86, 7, 20, 47
			' draw niche
			yImgDraw img(imgTabNext), 0, 0
			' draw next column
			yImgDraw blocks(nextCol(0) - minBlockId), 2, 3
			yImgDraw blocks(nextCol(1) - minBlockId), 2, 17
			yImgDraw blocks(nextCol(2) - minBlockId), 2, 31
		END IF

		IF (eng.dFlags AND efDwReactionTab) THEN
			' set viewport
			yViewport 61, 71, 47, 20

			'' FLASH COLUMNS (timer) ''
			IF (eng.gamMode) THEN
				' minutes
				tmpStr2 = LTRIM$(STR$(eng.tFlash \ 60))
				IF (LEN(tmpStr2) = 1) THEN
					tmpStr = "0" + tmpStr2 + ":"
				ELSE
					tmpStr = tmpStr2 + ":"
				END IF
				' seconds
				tmpStr2 = LTRIM$(STR$(eng.tFlash MOD 60))
				IF (LEN(tmpStr2) = 1) THEN
					tmpStr = tmpStr + "0" + tmpStr2
				ELSE
					tmpStr = tmpStr + tmpStr2
				END IF
				' draw
				tmp = 3
				FOR y% = 1 TO LEN(tmpStr)
					yImgDraw img(100 - 48 + ASC(MID$(tmpStr, y%, 1))), tmp, 2
					tmp = tmp + 8
				NEXT y%

			'' ORIGINAL MODE (reaction score) ''
			ELSE
				' clear tab
				yImgDraw img(imgTabReaction), 0, 0
				' print reaction score
				IF (eng.scoreTmp) THEN
					IF (eng.scoreTmp > 99999) THEN
						tmpStr = "99999"
					ELSE
						tmpStr = LTRIM$(STR$(eng.scoreTmp))
					END IF
					tmp = 3 + (5 - LEN(tmpStr)) * 8
					FOR y% = 1 TO LEN(tmpStr)
						yImgDraw img(100 - 48 + ASC(MID$(tmpStr, y%, 1))), tmp, 2
						tmp = tmp + 8
					NEXT y%
				END IF
			END IF
		END IF

		' reset viewport to grid
		yViewport 112, 9, 96, 182

		IF (eng.dFlags AND efDwLevelTab) THEN
			IF (eng.level > 999) THEN
				tmpStr = "999"
			ELSE
				tmpStr = LTRIM$(STR$(eng.level))
			END IF
			yPrint cMap(1), 13 - LEN(tmpStr), 18, tmpStr
		END IF

		IF (eng.dFlags AND efDwScoreTab) THEN
			'' FLASH COLUMNS (best time) ''
			IF (eng.gamMode) THEN
				' get best score for that configuration
				tmp = rank2(eng.skill + ((eng.floor - 2) * 3) + (eng.level * 24)).tmr
				' minutes
				tmpStr = LTRIM$(STR$(tmp \ 60)) + ":"
				' seconds
				tmpStr2 = LTRIM$(STR$(tmp MOD 60))
				IF (LEN(tmpStr2) = 1) THEN
					tmpStr = tmpStr + "0" + tmpStr2
				ELSE
					tmpStr = tmpStr + tmpStr2
				END IF
				' draw
				yPrint cMap(1), 13 - LEN(tmpStr), 14, tmpStr

			'' ORIGINAL MODE (score) ''
			ELSE
				IF (eng.score > 99999999) THEN
					tmpStr = "99999999"
				ELSE
					tmpStr = LTRIM$(STR$(eng.score))
				END IF
				yPrint cMap(1), 13 - LEN(tmpStr), 14, tmpStr
			END IF
		END IF

		IF (eng.dFlags AND efDwJewelsTab) THEN
			'' FLASH COLUMNS (skill) ''
			IF (eng.gamMode) THEN
				yPrint cMap(1), 10, 22, MID$("NOVAMAPRO", 1 + eng.skill * 3, 3)

			'' ORIGINAL MODE (jewels) ''
			ELSE
				' print gem count
				IF (eng.gems > 9999) THEN
					tmpStr = "9999"
				ELSE
					tmpStr = LTRIM$(STR$(eng.gems))
				END IF
				yPrint cMap(1), 13 - LEN(tmpStr), 22, tmpStr
			END IF
		END IF
	END IF

	'' gameplay grid
	IF (eng.dFlags AND efDwGridGems) THEN
		ofsY = 0
		ofsS = 17 ' start at 1 (because it's a string,) and add 2 rows of 8
		FOR y% = 2 TO 14 ' don't display top two and last rows
			ofsX = 0
			FOR x% = 1 TO 6 ' don't display leftmost and rightmost columns
				tmp = ASC(MID$(eng.grid, ofsS + x%, 1))
				IF (tmp <> ASC(MID$(eng.grid, ofsS + x% + 128, 1))) THEN
					yImgDraw img(imgCell), ofsX, ofsY ' clear
					SELECT CASE (tmp AND &H7)
					CASE minBlockId TO mgcBlockId
						yImgDraw blocks((tmp AND &H7) - minBlockId), ofsX, ofsY ' display gem
					END SELECT
				ELSEIF (tmp AND cellErase) THEN
					IF (eng.frames > 7) THEN
						yImgDraw img(43 + (eng.frames - 8)), ofsX, ofsY ' draw sprite
					ELSEIF (eng.frames AND 1) THEN
						yImgDraw blocks((tmp AND &H7) - minBlockId), ofsX, ofsY ' display
					ELSE
						yImgDraw img(imgCell), ofsX, ofsY ' clear
					END IF
				ELSEIF (tmp AND cellFlash) THEN
					yImgDraw img(imgCell), ofsX, ofsY ' clear
					yImgDraw blocks((tmp AND &H7) - minBlockId), ofsX, ofsY ' display
					IF (eng.glowFrame < 7) THEN yImgDraw img(imgGemGlow + eng.glowFrame), ofsX, ofsY ' draw glow
				END IF
				ofsX = ofsX + 16
			NEXT x%
			ofsY = ofsY + 14
			ofsS = ofsS + 8
		NEXT y%
	END IF

	SELECT CASE eng.turn
	CASE trGameOver, trWellDone
		FOR y% = 2 TO 14
			tmp = eng.frames - (22 - y%) ' offset lower than necessary
			SELECT CASE tmp
			CASE 0 TO 7 ' explosion
				ofsY = (y% - 2) * 14
				tmp = 43 + tmp
				yImgDraw img(tmp), 0, ofsY  ' explosion
				yImgDraw img(tmp), 16, ofsY ' explosion
				yImgDraw img(tmp), 32, ofsY ' explosion
				yImgDraw img(tmp), 48, ofsY ' explosion
				yImgDraw img(tmp), 64, ofsY ' explosion
				yImgDraw img(tmp), 80, ofsY ' explosion
			CASE IS > 7 ' erase
				yImgDraw img(imgCellRow), 0, (y% - 2) * 14
			END SELECT
		NEXT y%
		IF (eng.turn = trGameOver) THEN
			IF (eng.frames < 24) THEN
				yImgDraw img(imgGameOver), 4, 48 + (24 - eng.frames) * 10
			ELSE
				yImgDraw img(imgGameOver), 4, 48
			END IF
		ELSE
			IF (eng.frames < 24) THEN
				yImgDraw img(imgWellDone), 12, 48 + (24 - eng.frames) * 10
			ELSE
				yImgDraw img(imgWellDone), 12, 48
			END IF
		END IF
	
	CASE trPlayer
		IF ((eng.flags AND efTestGameOver) = 0) THEN
			ofsX = (eng.x - 1) * 16
			ofsY = (eng.y - 4) * 7
			yImgDraw blocks(thisCol(0) - minBlockId), ofsX, ofsY - 28
			yImgDraw blocks(thisCol(1) - minBlockId), ofsX, ofsY - 14
			yImgDraw blocks(thisCol(2) - minBlockId), ofsX, ofsY
		END IF

	CASE trStart
		yPrint cMap(2), 15, 12, "GAME START"

	CASE trInputName
		yPrint cMap(2), 15, 7, "VERY GOOD!"
		yImgDraw img(imgCellRow), 0, 5 * 14
		tmpStr = eng.plName
		IF (eng.flags AND efBlink) THEN MID$(tmpStr, eng.plNameOfs, 1) = " "
		yPrint cMap(2), 18, 10, tmpStr
	END SELECT

	'' print pause
	IF (eng.flags AND efPause) THEN yPrint cMap(2), 17, 7, "PAUSE!"

	'' shift for magic gem
	yPalShift 208, 231, 4
END SUB

''
'' Get PIT (Programmable Interval-Timer) countdown status, code by Seav (4660
'' PIT ticks = 1 second, 1 game tick = 78 PIT ticks)
''
FUNCTION getPIT& STATIC
	DIM Ticks AS LONG, LSB AS INTEGER, HSB AS INTEGER

	DEF SEG = 0
	Ticks = PEEK(&H46C) + PEEK(&H46D) * 256& + PEEK(&H46E) * 65536

	OUT &H43, &H4
	LSB = INP(&H40)
	HSB = 255 - INP(&H40)

	getPIT& = Ticks * 256& + HSB
END FUNCTION

''
'' Add player's column to the grid
''
SUB gridAdd STATIC
	DIM ofs AS INTEGER

	ofs = 1 + (eng.y \ 2) * 8 + eng.x
	MID$(eng.grid, ofs, 1) = CHR$(thisCol(2)): ofs = ofs - 8
	MID$(eng.grid, ofs, 1) = CHR$(thisCol(1)): ofs = ofs - 8
	MID$(eng.grid, ofs, 1) = CHR$(thisCol(0))
END SUB

''
'' Initialize grid
''
SUB gridClearAll
	DIM ofs AS INTEGER, tmp AS INTEGER, dummy AS STRING

	' reset
	ofs = 1
	FOR i% = 0 TO 14
		MID$(eng.grid, ofs, 8) = CHR$(cellVoid) + STRING$(6, 0) + CHR$(cellVoid)
		ofs = ofs + 8
	NEXT i%
	MID$(eng.grid, ofs, 8) = STRING$(8, cellVoid)

	' flash mode
	IF (eng.gamMode = 1) THEN
		' setup floor
		ofs = 1 + ((15 - eng.floor) * 8)
		tmp = 4 + eng.skill
		FOR i% = 0 TO eng.floor - 1
			MID$(eng.grid, ofs + 1, 6) = gridJunkRow$(tmp)
			ofs = ofs + 8
		NEXT i%
		' tag flash gem
		MID$(eng.grid, 117, 1) = CHR$(ASC(MID$(eng.grid, 117, 1)) OR cellFlash)
	END IF

	' force-reset mirror grid
	MID$(eng.grid, 129, 128) = STRING$(128, 255)
END SUB

''
'' "Random" grid junk, specify the number of pieces to obtain fresh junk.
'' The initial call performs some modifications; only get "new" junk at the
'' beginning of the game to ensure no automatic reaction can happen. Could be
'' used for a "high tide" mode in which new junk would be added to the field
'' over time.
''
'' Alternatively: for 5 blocks, use the 4-block list, pick an index and
'' replace at random that index by block id 5. The same can be done with 6
'' blocks (pick 4-block list, pick two indices, replace index A with block id
'' 5 at random, replace index B with block id 6 at random.)
''
FUNCTION gridJunkRow$ (numPieces AS INTEGER) STATIC
	DIM row AS INTEGER, col AS INTEGER, index(5) AS INTEGER
	DIM s AS STRING, t AS STRING, junkHeight AS INTEGER

	IF (numPieces) THEN
		' pick seed (40 lines, 6 columns each - 240 bytes)
		SELECT CASE numPieces
		CASE 4
			s = "311422142314233224441341331311224244113223113144"
			s = s + "324232144134332122412321331432424341421133313221"
			s = s + "344232121412423211212321434423241432131123244334"
			s = s + "231241313234212132131341421422313231424313242441"
			s = s + "243313112422214313432324112413243344443211212342"
		CASE 5
			s = "413542152341244225552452341311324255114334213125"
			s = s + "225443455135313242512231431543434351221134414321"
			s = s + "314343231513424211223221534553551542241123355445"
			s = s + "441331324545213232144341511523114332425414332552"
			s = s + "223314113513212423543314122413255455544311312553"
		CASE 6
			s = "513653163451244225662562551411223266114435224136"
			s = s + "326543566146443432613631441645535451121144125331"
			s = s + "415363642613535211223231645624361643241134365156"
			s = s + "352345454356214243255451611633215442536414332655"
			s = s + "263514213663332414645525122314465566655312312463"
		END SELECT

		' randomize indices
		index(0) = 1: index(1) = 2: index(2) = 3
		index(3) = 4: index(4) = 5: index(5) = 6
		FOR i% = 0 TO numPieces - 1
			SWAP index(i%), index(INT(RND * numPieces))
		NEXT i%

		' get junk height
		junkHeight = LEN(s) \ 6

		' re-index buffer
		FOR i% = 1 TO (junkHeight * 6)
			MID$(s, i%, 1) = CHR$(index(VAL(MID$(s, i%, 1)) - 1))
		NEXT i%

		' start at random row
		row = INT(RND * junkHeight) * 6

		' shift columns
		col = INT(RND * 6)
		IF (col) THEN
			FOR i% = 0 TO junkHeight - 1
				t = MID$(s, 1 + i% * 6, 6)
				MID$(s, 1 + i% * 6, 6) = RIGHT$(t, 6 - col) + LEFT$(t, col)
			NEXT i%
		END IF

		' reset value
		numPieces = 0
	END IF

	gridJunkRow$ = MID$(s, 1 + row, 6)
	row = (row + 6) MOD (junkHeight * 6)
END FUNCTION

''
'' remove gems tagged for deletion
''
SUB gridPrune STATIC
	DIM ofs AS INTEGER

	DEF SEG = VARSEG(eng.grid)
	ofs = VARPTR(eng.grid)

	' flash mode
	IF (eng.gamMode = 1) THEN
		' remove gems tagged for deletion
		eng.flags = eng.flags OR efFlashDone ' assume we're done
		FOR y% = 0 TO 127
			SELECT CASE (PEEK(ofs + y%) AND (cellFlash OR cellErase))
			CASE cellFlash
				eng.flags = eng.flags AND (NOT efFlashDone) ' we're not
			CASE IS > 0
				POKE (ofs + y%), 0
			END SELECT
		NEXT y%
	ELSE
		' remove gems tagged for deletion
		FOR y% = 0 TO 127
			IF (PEEK(ofs + y%) AND cellErase) THEN
				POKE (ofs + y%), 0
				eng.gems = eng.gems + 1
				eng.levelNext = eng.levelNext - 1
				eng.dFlags = eng.dFlags OR efDwJewelsTab
			END IF
		NEXT y%
		' gained at least one level
		IF (eng.levelNext < 1) THEN
			WHILE (eng.levelNext < 1)
				eng.levelNext = eng.levelNext + 50
				eng.level = eng.level + 1
				eng.levelPts = (eng.level + 1) * 10
				eng.pushDelay = eng.pushDelay - 10
				IF (eng.pushDelay < 4) THEN eng.pushDelay = 4
				IF ((eng.level AND &H3) = 0) THEN yPalShift 16, 31, 8
			WEND
			eng.dFlags = eng.dFlags OR efDwLevelTab
			playSound sndLevel
		END IF
	END IF
END SUB

''
'' stack gems down
''
FUNCTION gridStack% STATIC
	DIM ofsS AS INTEGER, ofs AS INTEGER, tmp AS INTEGER, max AS INTEGER

	DEF SEG = VARSEG(eng.grid)
	ofs = VARPTR(eng.grid)

	' stack
	gridStack% = 0
	FOR y% = 14 TO 1 STEP -1
		ofsS = ofs + y% * 8
		FOR x% = 1 TO 6
			IF PEEK(ofsS + x%) = 0 THEN
				tmp = PEEK(ofsS + x% - 8)
				SELECT CASE tmp
				CASE minBlockId TO maxBlockId
					POKE ofsS + x% - 8, 0
					POKE ofsS + x%, tmp
					gridStack% = -1
				END SELECT
			END IF
		NEXT x%
	NEXT y%
END FUNCTION

''
'' TEST WHOLE GRID FOR STREAKS OF 3 OR MORE GEMS
''
SUB gridTest1 STATIC
	DIM tmpPrev AS INTEGER, coords(3) AS INTEGER, numCells AS INTEGER
	DIM ofsS AS INTEGER, tmp AS INTEGER, aligned AS INTEGER, gotMatch AS INTEGER

	' work with a copy of the grid
	memCpy VARSEG(eng.grid), VARPTR(eng.grid), eng.sListSeg, 0, 128

	' search list (horizontal, vertical, diagonal, counter-diagonal)
	gotMatch = 0
	DEF SEG = eng.sListSeg
	ofsS = 128
	DO
		' new search
		tmpPrev = -1
		numCells = PEEK(ofsS)

		' visit each address for this search
		FOR i% = 1 TO numCells
			' preserve last three addresses
			coords(2) = coords(1)
			coords(1) = coords(0)
			coords(0) = PEEK(ofsS + i%)

			' what's in that cell?
			tmp = PEEK(coords(0)) AND &H7

			' not a gem, or a different gem: cancel streak
			IF (tmp < minBlockId) OR (tmp > maxBlockId) OR (tmp <> tmpPrev) THEN
				aligned = 1

			' same gem as previous, increase streak
			ELSE
				aligned = aligned + 1
				SELECT CASE aligned
				CASE 3 ' got 3? delete!
					POKE coords(2), (PEEK(coords(2)) OR cellErase)
					POKE coords(1), (PEEK(coords(1)) OR cellErase)
					POKE coords(0), (PEEK(coords(0)) OR cellErase)
					gotMatch = gotMatch + 3
				CASE IS > 3 ' delete all subsequent gems
					POKE coords(0), (PEEK(coords(0)) OR cellErase)
					gotMatch = gotMatch + 1
				END SELECT
			END IF

			' keep track of this gem
			tmpPrev = tmp
		NEXT i%

		' move to next search
		ofsS = ofsS + numCells + 1
	LOOP WHILE (ofsS < eng.sListLen)

	' update game grid
	memCpy eng.sListSeg, 0, VARSEG(eng.grid), VARPTR(eng.grid), 128

	' we got a match, stack gems
	IF (gotMatch) THEN
		' increase score (original mode only)
		IF (eng.gamMode = 0) THEN
			eng.scoreTmp = eng.scoreTmp + (gotMatch * eng.levelPts)
		END IF
		' set turn to explosion
		eng.countdown = eng.exploDelay
		eng.turn = trExplosion
		eng.dFlags = eng.dFlags OR efDwReactionTab
		eng.chainReact = eng.chainReact + 1
		playSound sndExplo
	END IF
END SUB

''
'' TEST FOR MAGIC GEM
''
SUB gridTest2 STATIC
	DIM idBlock AS INTEGER, tmp AS INTEGER, ofs AS INTEGER

	DEF SEG = VARSEG(eng.grid)
	ofs = VARPTR(eng.grid)

	' destroy magic column
	tmp = eng.x + (eng.y \ 2) * 8
	POKE (ofs + tmp - 16), mgcBlockId OR cellErase
	POKE (ofs + tmp - 8), mgcBlockId OR cellErase
	POKE (ofs + tmp), mgcBlockId OR cellErase

	' break similar blocks
	idBlock = PEEK(ofs + tmp + 8)
	IF (idBlock <> cellVoid) THEN
		FOR y% = 0 TO 127
			IF (PEEK(ofs + y%) = idBlock) THEN
				POKE ofs + y%, idBlock OR cellErase
				eng.scoreTmp = eng.scoreTmp + eng.levelPts
			END IF
		NEXT y%
	END IF

	' stack gems
	eng.countdown = eng.exploDelay
	eng.turn = trExplosion
	eng.dFlags = eng.dFlags OR efDwReactionTab
	eng.chainReact = eng.chainReact + 1
	playSound sndExplo2
END SUB

SUB hiscoreLoop
	DIM tPrevious AS LONG, tLag AS LONG, tCurrent AS LONG, tElapsed AS LONG
	DIM lst(8) AS STRING * 25, tmp AS STRING, tmpInt1 AS INTEGER, tmpInt2 AS INTEGER
	DIM frame AS INTEGER, anim AS INTEGER, countdown AS INTEGER, oldFrame AS INTEGER

	' convert score to string
	FOR i% = 0 TO 8
		MID$(lst(i%), 1, 1) = CHR$(49 + i%)
		MID$(lst(i%), 3, 3) = LEFT$(rank1(i%).pl, 3)
		tmp = LTRIM$(STR$(rank1(i%).score)): MID$(lst(i%), 15 - LEN(tmp), LEN(tmp)) = tmp
		tmp = LTRIM$(STR$(rank1(i%).jewels)): MID$(lst(i%), 20 - LEN(tmp), LEN(tmp)) = tmp
		tmp = LTRIM$(STR$(rank1(i%).level)): MID$(lst(i%), 26 - LEN(tmp), LEN(tmp)) = tmp
	NEXT i%

	eng.dFlags = efDwHiScore

	tPrevious = getPIT&
	DO
		' update timer
		tCurrent = getPIT&
		tElapsed = tCurrent - tPrevious
		tPrevious = tCurrent
		tLag = tLag + tElapsed

		' update game logic for 78 clock ticks for every
		' slice of 78 ticks we skipped during rendering.
		WHILE (tLag > 78)

			' capture keys
			commonProcess

			countdown = countdown + (countdown > 0)

			IF (eng.keyCD = 0) THEN
				IF (eng.keyCtrl AND cAction) THEN
					eng.flags = eng.flags OR efEscape
				END IF
			END IF

			IF (countdown = 0) THEN
				SELECT CASE anim
				CASE 0 ' show score
					IF (frame < 50) THEN
						countdown = 2
						frame = frame + 1
					ELSE
						anim = 1 ' yeah... no. I'd need to rewrite huge chunks of the
						' library to have the animated gems out of the purse like I want
						' to. I'll wrap it up for now.
					END IF
				END SELECT
			END IF

			' A game tick is 78 PIT ticks, 4660 PIT ticks = 1 second
			tLag = tLag - 78
		WEND

		IF (eng.dFlags AND efDwHiScore) THEN
			yViewport 0, 0, 320, 200
			yScreenClear 81             ' clear screen
			FOR i% = 0 TO 319 STEP 16
				yImgDraw img(88), i%, 0   ' top trim
				yImgDraw img(88), i%, 192 ' bottom trim
			NEXT i%
			yImgDraw img(80), 281, 62   ' right pillar
			yImgDraw img(86), 280, 8    ' statue
			yImgDraw img(84), 272, 56   ' tall fern (sprite)
			yImgDraw img(81), 273, 142  ' short fern (sprite)
			yImgDraw img(82), 306, 142  ' short fern (sprite)
			yPrint cMap(2), 10, 3, "SCORE JEWELS LEVEL"
		END IF

		IF (frame <> oldFrame) THEN
			oldFrame = frame
			tmpInt1 = frame
			tmpInt2 = 5
			FOR i% = 0 TO 8
				IF (tmpInt1 < 1) THEN EXIT FOR
				yPrintColor cMap(1), 81, 2, tmpInt2, STRING$(25, "A")
				yPrint cMap(2), 2, tmpInt2, RIGHT$(lst(i%), tmpInt1)
				tmpInt1 = tmpInt1 - 3
				tmpInt2 = tmpInt2 + 2
			NEXT i%
		END IF

		yFlip
		yScreenCopy
		IF (eng.flags AND efScreenshot) THEN
			screenshot
			eng.flags = eng.flags AND (NOT efScreenshot)
		END IF

		eng.dFlags = 0

	LOOP UNTIL (eng.flags AND efEscape)
	eng.flags = eng.flags AND (NOT efEscape)
END SUB

''
'' Load files, set palette, etc.
''
SUB initProg
	DIM tmpStr AS STRING, tmpInt AS INTEGER

	' you know what this is
	RANDOMIZE TIMER

	' switch to mode Y
	yScreen -1

	' ---- LOAD DATA FOR SPLASH SCREEN ---- '
	img(89) = yImgFromFile%("DATA/PHONEY.BSV", vgaSolid)  ' logo
	img(61) = yImgFromFile%("DATA/CMAP2.BSV", vgaSprite)  ' character map
	cMap(0) = yCMapSet%(img(61), vgaSolid, 33, 122)       ' configure cMap

	tmpInt = yPalFromFile%("DATA/PALETTE.ACT")            ' load palette
	yPalApply tmpInt, 0, 256                              ' apply right away
	yPalFree tmpInt                                       ' free palette

	' display splash screen while we're loading the rest of the resources...
	yImgDraw img(89), 96, 56
	SELECT CASE INT(RND * 7)
	CASE 0
		tmpStr = "Written entirely in under an hour!"
	CASE 1
		tmpStr = "I'm broke, please don't sue me."
	CASE 2
		tmpStr = "DOS does what Windidn't. Or something."
	CASE 3
		tmpStr = "Columns me your color, baby. Columns me your car."
	CASE 4
		tmpStr = "It's more about gems and less about holograms."
	CASE 5
		tmpStr = "A division of Super Legit Global Corp."
	CASE 6
		tmpStr = "We got chain reactions!"
	END SELECT
	yPrint cMap(0), 40 - LEN(tmpStr) \ 2, 13, tmpStr
	yFlip

	' load backdrop
	img(imgBlock1) = yImgFromFile%("DATA/BLOCK1.BSV", vgaSolid)
	img(imgBlock2) = yImgFromFile%("DATA/BLOCK2.BSV", vgaSolid)
	img(imgCell) = yImgFromFile%("DATA/CELL.BSV", vgaSolid)
	img(imgTabNext) = yImgFromFile%("DATA/TAB_NEXT.BSV", vgaSolid)
	img(imgTabReaction) = yImgFromFile%("DATA/TAB_REAC.BSV", vgaSolid)
	img(imgTabScore) = yImgFromFile%("DATA/TAB_SCRE.BSV", vgaSolid)
	img(imgTabLevel) = yImgFromFile%("DATA/TAB_LEVL.BSV", vgaSolid)
	img(imgTabJewels) = yImgFromFile%("DATA/TAB_JEWL.BSV", vgaSolid)
	img(imgCellRow) = yImgFromFile%("DATA/CELLROW.BSV", vgaSolid)
	img(imgGameOver) = yImgFromFile%("DATA/GAMEOVER.BSV", vgaSprite)
	img(imgWellDone) = yImgFromFile%("DATA/WELLDONE.BSV", vgaSprite)
	img(imgTabBest) = yImgFromFile%("DATA/TAB_BEST.BSV", vgaSolid)
	img(imgTabSkill) = yImgFromFile%("DATA/TAB_CLSS.BSV", vgaSolid)

	' load gems
	FOR i% = 0 TO 5
		tmpStr = CHR$(48 + i%) + ".BSV"
		img(imgJems + i%) = yImgFromFile%("DATA/GEMS" + tmpStr, vgaSprite)
		img(imgJems + 6 + i%) = yImgFromFile%("DATA/DICE" + tmpStr, vgaSprite)
		img(imgJems + 12 + i%) = yImgFromFile%("DATA/SQUARES" + tmpStr, vgaSprite)
		img(imgJems + 18 + i%) = yImgFromFile%("DATA/CARDS" + tmpStr, vgaSprite)
		img(imgJems + 24 + i%) = yImgFromFile%("DATA/FRUITS" + tmpStr, vgaSprite)
	NEXT i%

	' super gem
	img(40) = yImgFromFile%("DATA/MAGIC0.BSV", vgaSprite)

	' load explosion
	FOR i% = 0 TO 7
		img(43 + i%) = yImgFromFile%("DATA/EXPLO" + CHR$(48 + i%) + ".BSV", vgaSolid)
	NEXT i%

	img(51) = yImgFromFile%("DATA/EDGE1.BSV", vgaSolid)
	img(52) = yImgFromFile%("DATA/EDGE2.BSV", vgaSolid)
	img(53) = yImgFromFile%("DATA/EDGE3.BSV", vgaSolid)
	img(54) = yImgFromFile%("DATA/EDGE4.BSV", vgaSolid)

	' load standard font
	img(60) = yImgFromFile%("DATA/CMAP1.BSV", vgaSprite) ' Columns font
	cMap(1) = yCMapSet%(img(60), vgaSolid, 33, 91)       ' solid (game UI)
	cMap(2) = yCMapSet%(img(60), vgaSprite, 33, 91)      ' transparent (in menu)

	' load wide characters
	FOR i% = 0 TO 9
		img(100 + i%) = yImgFromFile%("DATA/NUM" + CHR$(48 + i%) + ".BSV", vgaSolid)
	NEXT i%
	img(110) = yImgFromFile%("DATA/NUMCOLON.BSV", vgaSolid)

	' load menu graphics (not in fast mode)
	img(80) = yImgFromFile%("DATA/_COLUMN.BSV", vgaSolid)
	img(81) = yImgFromFile%("DATA/_FERNA1.BSV", vgaSprite)
	img(82) = yImgFromFile%("DATA/_FERNA2.BSV", vgaSprite)
	img(83) = yImgFromFile%("DATA/_FERNB1.BSV", vgaSprite)
	img(84) = yImgFromFile%("DATA/_FERNB2.BSV", vgaSprite)
	img(85) = yImgFromFile%("DATA/_STATLF.BSV", vgaSolid)
	img(86) = yImgFromFile%("DATA/_STATRT.BSV", vgaSolid)
	img(87) = yImgFromFile%("DATA/_TITLE.BSV", vgaSolid)
	img(88) = yImgFromFile%("DATA/_TRIM.BSV", vgaSolid)

	' block previews
	FOR i% = 0 TO 4
		img(90 + i%) = yImgFromFile%("DATA/_PREV" + CHR$(48 + i%) + ".BSV", vgaSprite)
		cMap(3 + i%) = yCMapSet%(img(90 + i%), vgaSprite, 0, 5)
	NEXT i%

	' glowing effect
	FOR i% = 0 TO 6
		img(imgGemGlow + i%) = yImgFromFile%("DATA/GLOW" + CHR$(48 + i%) + ".BSV", vgaSprite)
	NEXT i%

	' load music tracks
	music(0) = loadTrack$("DATA/SELECT.SPK")
	music(1) = loadTrack$("DATA/START.SPK")
	music(2) = loadTrack$("DATA/GAMEOVER.SPK")
	music(3) = loadTrack$("DATA/ENTRY.SPK")
	music(4) = loadTrack$("DATA/CLOTHO.SPK"): musicFlag(4) = msLoop
	music(5) = loadTrack$("DATA/ATROPOS.SPK"): musicFlag(5) = msLoop
	music(6) = loadTrack$("DATA/LATHESIS.SPK"): musicFlag(6) = msLoop

	' set defaults
	eng.plTrack = 4
	eng.floor = 2

	' default playback mode
	PLAY "MN MB"

	' search list (for gem matches)
	OPEN "DATA/SEARCH.LST" FOR BINARY AS #1
	tmpStr = SPACE$(LOF(1))
	GET #1, , tmpStr
	CLOSE #1
	eng.sListLen = 128 + LEN(tmpStr)
	eng.sListSeg = memAlloc%(CLNG(eng.sListLen))
	memCpy VARSEG(tmpStr), SADD(tmpStr), eng.sListSeg, 128, LEN(tmpStr)

	' load scores
	initScores
	IF fileExists%("SCORES.HI") THEN
		OPEN "SCORES.HI" FOR BINARY AS #1
		FOR i% = 0 TO 8
			GET #1, , rank1(i%)
		NEXT i%
		CLOSE #1
	END IF
	IF fileExists%("FLASH.HI") THEN
		OPEN "FLASH.HI" FOR BINARY AS #1
		FOR i% = 0 TO 239
			GET #1, , rank2(i%)
		NEXT i%
		CLOSE #1
	END IF
END SUB

''
'' Set default scores
''
SUB initScores
	' original mode highscore table
	FOR i% = 0 TO 8
		rank1(i%).score = (9 - i%) * 1000
		rank1(i%).jewels = (9 - i%) * 10
		rank1(i%).level = (9 - i%)
	NEXT i%
	rank1(0).score = 10000: rank1(1).score = 9000
	rank1(0).pl = "WNK": rank1(1).pl = "AIA": rank1(2).pl = "KMZ"
	rank1(3).pl = "III": rank1(4).pl = "HYM": rank1(5).pl = "AAO"
	rank1(6).pl = "R.T": rank1(7).pl = "AHO": rank1(8).pl = "???"
	' best time
	FOR i% = 0 TO 239
		rank2(i%).pl = "???"
		rank2(i%).tmr = 0
	NEXT i%
END SUB

''
'' New keyboard handler, returns segment where key status are stored. (DOS 2+)
''
FUNCTION keyboardInit% (mode AS INTEGER) STATIC
	DIM oldISRSeg AS INTEGER, oldISROfs AS INTEGER, newISRSeg AS INTEGER
	DIM regs AS RegTypeX

	IF (mode = kbEnable) THEN
		IF (newISRSeg = 0) THEN
			' reserve memory for buffer & code
			newISRSeg = memAlloc%(169) ' key status buffer (129) + code (40)

			' write code
			DEF SEG = newISRSeg
			FOR i% = 0 TO 39
				POKE i% + 129, VAL("&H" + MID$("9C5053E460B401A8807404B400247F88C3B7002E8827E4610C80E661247FE661B020E6205B589DCF", 1 + i% * 2, 2))
			NEXT i%

			' preserve vector interrupt 9 (BIOS keyboard ISR)
			regs.ax = &H3509
			CALL INTERRUPTX(&H21, regs, regs)
			oldISRSeg = regs.es
			oldISROfs = regs.bx
		END IF

		' clear key strokes
		DEF SEG = newISRSeg
		FOR i% = 0 TO 128
			POKE i%, 0
		NEXT i%

		' Clear keyboard buffer
		DEF SEG = 0
		POKE (&H41A), PEEK(&H41C)

		' hook custom keyboard handler
		regs.ax = &H2509
		regs.ds = newISRSeg
		regs.DX = 129
		CALL INTERRUPTX(&H21, regs, regs)

		' return offset to key status buffer
		keyboardInit% = newISRSeg
	ELSE
		IF (newISRSeg) THEN
			' restore BIOS keyboard ISR
			regs.ax = &H2509
			regs.ds = oldISRSeg
			regs.DX = oldISROfs
			CALL INTERRUPTX(&H21, regs, regs)
		END IF

		' deallocate memory reserved for buffer & code
		IF (mode = kbUninstall) THEN memFree newISRSeg

		' mark as disabled (memory not be freed)
		keyboardInit% = 0
	END IF
END FUNCTION

''
'' Load PC Speaker tracks
''
FUNCTION loadTrack$ (filename AS STRING)
	DIM ff AS INTEGER, tmp AS STRING
	ff = FREEFILE

	IF (fileExists%(filename)) THEN
		OPEN filename FOR INPUT AS #ff
		DO
			LINE INPUT #ff, tmp
		LOOP WHILE (ASC(LEFT$(tmp, 1)) = 39)
		CLOSE #ff
	ELSE
		tmp = " "
	END IF

	loadTrack$ = tmp
END FUNCTION

''
'' MENU LOOP
''
FUNCTION menuLoop%
	DIM tPrevious AS LONG, tLag AS LONG, tCurrent AS LONG, tElapsed AS LONG
	DIM msg AS STRING * 8, tmp AS INTEGER

	DIM menuOffset1 AS INTEGER, menuOffset2 AS INTEGER
	DIM itemCount AS INTEGER, itemSelect AS INTEGER
	DIM it(7) AS itemType

	' ticks between identical key presses
	eng.keyDelay = 15

	' initialize main menu, force redraw
	eng.dFlags = efDwMenu
	'eng.menuIndex = 0
	GOSUB menuLoopSwap

	' play title theme
	eng.track = 0

	tPrevious = getPIT&
	DO
		' update timer
		tCurrent = getPIT&
		tElapsed = tCurrent - tPrevious
		tPrevious = tCurrent
		tLag = tLag + tElapsed

		' update game logic for 78 clock ticks for every
		' slice of 78 ticks we skipped during rendering.
		WHILE (tLag > 78)

			' capture keys
			commonProcess

			IF (eng.keyCD = 0) THEN
				IF (eng.keyCtrl AND cAction) THEN
					IF (it(itemSelect).flags AND mfChangeMenu) THEN ' change menu
						eng.menuIndex = (it(itemSelect).flags \ &H100)
						GOSUB menuLoopSwap
					ELSEIF (it(itemSelect).flags AND mfExitMenu) THEN ' start game or quit to DOS
						eng.flags = eng.flags OR efEscape
						menuLoop% = (it(itemSelect).flags \ &H100)
					ELSEIF (it(itemSelect).flags AND mfClearScores) THEN
						initScores
						eng.menuIndex = (it(itemSelect).flags \ &H100)
						GOSUB menuLoopSwap
					END IF
				END IF

				' up or down
				IF (eng.keyCtrl AND cUp) THEN
					tmp = itemSelect
					DO
						itemSelect = (itemSelect + itemCount - 1) MOD itemCount
					LOOP WHILE ((it(itemSelect).flags AND (NOT mfScore)) = 0)
					IF (itemSelect <> tmp) THEN playSound sndShift
				ELSEIF (eng.keyCtrl AND cDown) THEN
					tmp = itemSelect
					DO
						itemSelect = (itemSelect + 1) MOD itemCount
					LOOP WHILE ((it(itemSelect).flags AND (NOT mfScore)) = 0)
					IF (itemSelect <> tmp) THEN playSound sndShift
				END IF

				' left or right
				IF (eng.keyCtrl AND cLeft) THEN
					IF (it(itemSelect).flags AND mfCounter) THEN
						SELECT CASE (it(itemSelect).flags \ &H100)
						CASE 1 ' skill
							eng.skill = eng.skill + (eng.skill > 0)
							FOR i% = 0 TO itemCount - 1
								IF (((it(i%).flags AND mfCounter) AND (it(i%).flags \ &H100 = 4)) OR ((it(i%).flags AND mfScore) <> 0)) THEN
									it(i%).flags = it(i%).flags OR mfRedraw
								END IF
							NEXT i%
						CASE 2 ' level
							eng.levelStart = eng.levelStart + (eng.levelStart > 0)
							FOR i% = 0 TO itemCount - 1
								IF ((it(i%).flags AND mfScore) <> 0) THEN
									it(i%).flags = it(i%).flags OR mfRedraw
								END IF
							NEXT i%
						CASE 3 ' music
							eng.plTrack = eng.plTrack + (eng.plTrack > 4)
						CASE 4 ' blocks
							eng.blockType = eng.blockType + (eng.blockType > 0)
						CASE 5 ' floor
							eng.floor = eng.floor + (eng.floor > 2)
							FOR i% = 0 TO itemCount - 1
								IF (it(i%).flags AND mfScore) THEN it(i%).flags = it(i%).flags OR mfRedraw
							NEXT i%
						CASE 6 ' sound (music/sfx/none)
							eng.sndMode = eng.sndMode + (eng.sndMode > 0)
						CASE 7 ' controls
							eng.flags = eng.flags AND (NOT efJoystick)
						END SELECT
						it(itemSelect).flags = it(itemSelect).flags OR mfRedraw
						eng.dFlags = eng.dFlags OR efDwMenuItems
					END IF
				ELSEIF (eng.keyCtrl AND cRight) THEN
					IF (it(itemSelect).flags AND mfCounter) THEN
						SELECT CASE (it(itemSelect).flags \ &H100)
						CASE 1 ' skill
							eng.skill = eng.skill - (eng.skill < 2)
							FOR i% = 0 TO itemCount - 1
								IF (((it(i%).flags AND mfCounter) AND (it(i%).flags \ &H100 = 4)) OR ((it(i%).flags AND mfScore) <> 0)) THEN
									it(i%).flags = it(i%).flags OR mfRedraw
								END IF
							NEXT i%
						CASE 2 ' level
							eng.levelStart = eng.levelStart - (eng.levelStart < 9)
							FOR i% = 0 TO itemCount - 1
								IF (it(i%).flags AND mfScore) THEN it(i%).flags = it(i%).flags OR mfRedraw
							NEXT i%
						CASE 3 ' music
							eng.plTrack = eng.plTrack - (eng.plTrack < 6)
						CASE 4 ' blocks
							eng.blockType = eng.blockType - (eng.blockType < 4)
						CASE 5 ' floor
							eng.floor = eng.floor - (eng.floor < 9)
							FOR i% = 0 TO itemCount - 1
								IF (it(i%).flags AND mfScore) THEN it(i%).flags = it(i%).flags OR mfRedraw
							NEXT i%
						CASE 6 ' sound (sfx/music/none)
							eng.sndMode = eng.sndMode - (eng.sndMode < 2)
						CASE 7 ' controls
							eng.flags = eng.flags OR efJoystick
						END SELECT
						it(itemSelect).flags = it(itemSelect).flags OR mfRedraw
						eng.dFlags = eng.dFlags OR efDwMenuItems
					END IF
				END IF
			END IF

		' A game tick is 78 PIT ticks, 4660 PIT ticks = 1 second
			tLag = tLag - 78
		WEND

		'' RENDERING, display next frame ''
		' draw background
		IF (eng.dFlags AND efDwMenu) THEN
			yViewport 0, 0, 320, 200    ' reset viewport
			yScreenClear 81             ' clear screen
			FOR i% = 0 TO 319 STEP 16
				yImgDraw img(88), i%, 0   ' top trim
				yImgDraw img(88), i%, 192 ' bottom trim
			NEXT i%
			yImgDraw img(87), 48, 8     ' big title
			yImgDraw img(80), 9, 62     ' left pillar
			yImgDraw img(85), 8, 8      ' statue
			yImgDraw img(83), 33, 56    ' tall fern (sprite)
			yImgDraw img(81), 1, 142    ' short fern (sprite)
			yImgDraw img(82), 34, 142   ' short fern (sprite)
			yImgDraw img(80), 281, 62   ' right pillar
			yImgDraw img(86), 280, 8    ' statue
			yImgDraw img(84), 272, 56   ' tall fern (sprite)
			yImgDraw img(81), 273, 142  ' short fern (sprite)
			yImgDraw img(82), 306, 142  ' short fern (sprite)
		END IF
		' draw menus
		IF (eng.dFlags AND efDwMenuItems) THEN
			' clear menu space
			IF (eng.dFlags AND efDwMenuClear) THEN
				FOR i% = 12 TO 21
					yPrintColor cMap(1), 81, 9, i%, "AAAAAAAAAAAAAAAAAAAAAA"
				NEXT i%
			END IF
			FOR i% = 0 TO itemCount - 1
				IF (it(i%).flags AND mfRedraw) THEN
					it(i%).flags = it(i%).flags XOR mfRedraw
					yPrintColor cMap(1), 81, menuOffset1 + 2, it(i%).y, "AAAAAAAAAAAAAAAAAAAAAA"
					yPrint cMap(2), menuOffset1 + 2, it(i%).y, RTRIM$(it(i%).msg)
					IF (it(i%).flags AND mfCounter) THEN
						IF ((it(i%).flags \ &H100) = 4) THEN ' blocks
							msg = LEFT$(CHR$(0) + CHR$(1) + CHR$(2) + CHR$(3) + CHR$(4) + CHR$(5), 4 + eng.skill)
							yPrint cMap(3 + eng.blockType), menuOffset2, it(i%).y, RTRIM$(LTRIM$(msg))
						ELSE
							SELECT CASE (it(i%).flags \ &H100)
							CASE 1 ' skill
								msg = MID$("NOVICE  AMATEUR PRO     ", 1 + eng.skill * 8, 8)
							CASE 2 ' level
								msg = STR$(eng.levelStart)
							CASE 3 ' music
								msg = MID$("CLOTHO  ATROPOS LATHESIS", 1 + (eng.plTrack - 4) * 8, 8)
							CASE 5 ' floor
								msg = STR$(eng.floor)
							CASE 6 ' sound (music/sfx/none)
								msg = MID$("EFFECTS MUSIC   NONE    ", 1 + eng.sndMode * 8, 8)
							CASE 7 ' controls
								IF (eng.flags AND efJoystick) THEN
									msg = "JOYSTICK"
								ELSE
									msg = "KEYBOARD"
								END IF
							END SELECT
							yPrintColor cMap(2), iColor, menuOffset2, it(i%).y, RTRIM$(LTRIM$(msg))
						END IF
					ELSEIF (it(i%).flags AND mfScore) THEN
						tmp = eng.skill + ((eng.floor - 2) * 3) + (eng.levelStart * 24)
						yPrint cMap(2), menuOffset1 + 3 + LEN(RTRIM$(it(i%).msg)), it(i%).y, LTRIM$(STR$(rank2(tmp).tmr \ 60)) + ":" + RIGHT$("0" + LTRIM$(STR$(rank2(tmp).tmr MOD 60)), 2) + " BY " + LEFT$(rank2(tmp).pl, 3)
					END IF
				END IF
			NEXT i%
		END IF
		' draw cursor
		yPrint cMap(2), menuOffset1, it(itemSelect).y, ">"
		'yPrintColor cMap(2), iColor, menuOffset1, it(itemSelect).y, ">"
		yFlip
		IF (eng.flags AND efScreenshot) THEN
			screenshot
			eng.flags = eng.flags AND (NOT efScreenshot)
		END IF
		' ready back buffer
		yScreenCopy
		yPrintColor cMap(2), 81, menuOffset1, it(itemSelect).y, ">"

		' reset draw flags
		eng.dFlags = 0
	LOOP UNTIL (eng.flags AND efEscape)
	eng.flags = eng.flags AND (NOT efEscape)
	EXIT FUNCTION

menuLoopSwap:
	' reset cursor location
	itemSelect = 0
	menuOffset1 = 10

	' set default row for each item
	FOR i% = 0 TO UBOUND(it)
		it(i%).y = 12 + i%
		it(i%).flags = 0
	NEXT i%

	' init menu
	SELECT CASE eng.menuIndex
	CASE 0
		itemCount = 6
		it(0).msg = "ORIGINAL MODE": it(0).flags = (3 * &H100) OR mfChangeMenu
		it(1).msg = "COLUMNS FLASH": it(1).flags = (4 * &H100) OR mfChangeMenu
		it(2).msg = "HIGHSCORES": it(2).flags = (1 * &H100) OR mfExitMenu
		it(3).msg = "OPTIONS": it(3).flags = (2 * &H100) OR mfChangeMenu
		it(4).msg = "ABOUT": it(4).flags = (1 * &H100) OR mfChangeMenu
		it(5).msg = "QUIT TO DOS": it(5).y = 18: it(5).flags = (6 * &H100) OR mfChangeMenu
	CASE 1
		itemCount = 5
		itemSelect = 4
		menuOffset1 = 8
		it(0).msg = "  COLUMNS IS @ SEGA"
		it(1).msg = "ADDITIONAL GRAPHICS &": it(1).y = 14
		it(2).msg = "QUICKBASIC VERSION BY": it(2).y = 15
		it(3).msg = "      MIKE HAWK": it(3).y = 16
		it(4).msg = "RETURN": it(4).y = 19: it(4).flags = (0 * &H100) OR mfChangeMenu
	CASE 2
		itemCount = 4
		menuOffset2 = 21
		it(0).msg = "CONTROLS": it(0).flags = (7 * &H100) OR mfCounter
		it(1).msg = "SOUND": it(1).flags = (6 * &H100) OR mfCounter
		it(2).msg = "ERASE SCORES": it(2).flags = (5 * &H100) OR mfChangeMenu
		it(3).msg = "RETURN": it(3).y = 16: it(3).flags = (0 * &H100) OR mfChangeMenu
	CASE 3
		itemCount = 6
		menuOffset2 = 19
		it(0).msg = "SKILL": it(0).flags = (1 * &H100) OR mfCounter
		it(1).msg = "LEVEL": it(1).flags = (2 * &H100) OR mfCounter
		it(2).msg = "MUSIC": it(2).flags = (3 * &H100) OR mfCounter
		it(3).msg = "BLOCKS": it(3).flags = (4 * &H100) OR mfCounter
		it(4).msg = "GO!": it(4).y = 17: it(4).flags = (2 * &H100) OR mfExitMenu ' original
		it(5).msg = "RETURN": it(5).y = 18: it(5).flags = (0 * &H100) OR mfChangeMenu
	CASE 4
		itemCount = 8
		menuOffset2 = 19
		it(0).msg = "SKILL": it(0).flags = (1 * &H100) OR mfCounter
		it(1).msg = "LEVEL": it(1).flags = (2 * &H100) OR mfCounter
		it(2).msg = "FLOOR": it(2).flags = (5 * &H100) OR mfCounter
		it(3).msg = "MUSIC": it(3).flags = (3 * &H100) OR mfCounter
		it(4).msg = "BLOCKS": it(4).flags = (4 * &H100) OR mfCounter
		it(5).msg = "BEST: ": it(5).y = 18: it(5).flags = mfScore
		it(6).msg = "GO!": it(6).y = 20: it(6).flags = (3 * &H100) OR mfExitMenu ' flash
		it(7).msg = "RETURN": it(7).y = 21: it(7).flags = (0 * &H100) OR mfChangeMenu
	CASE 5
		itemCount = 4
		itemSelect = 3
		it(0).msg = "  ERASE SCORES"
		it(1).msg = "  ARE YOU SURE?"
		it(2).msg = "YES": it(2).y = 15: it(2).flags = (2 * &H100) OR mfClearScores ' clear score, go back to menu 2
		it(3).msg = "CANCEL!": it(3).y = 16: it(3).flags = (2 * &H100) OR mfChangeMenu
	CASE 6
		itemCount = 4
		itemSelect = 3
		it(0).msg = "   QUIT TO DOS"
		it(1).msg = "  ARE YOU SURE?"
		it(2).msg = "YES": it(2).y = 15: it(2).flags = (0 * &H100) OR mfExitMenu
		it(3).msg = "NO": it(3).y = 16: it(3).flags = (0 * &H100) OR mfChangeMenu
	END SELECT

	' tag all for redraw
	eng.dFlags = eng.dFlags OR efDwMenuClear OR efDwMenuItems
	FOR i% = 0 TO itemCount - 1
		it(i%).flags = it(i%).flags OR mfRedraw
	NEXT i%
	RETURN
END FUNCTION

''
'' PLAY SOUND EFFECT
''
SUB playSound (id AS INTEGER)
	IF (eng.sndMode <> sndEffect) THEN EXIT SUB
	IF (PLAY(0) <> 0) AND (id <> sndExplo) THEN EXIT SUB

	PLAY "T120"
	SELECT CASE id
	CASE sndShift
		PLAY "L32O4C#A#16"
	CASE sndPlace
		PLAY "O0G16G32"
	CASE sndExplo
		PLAY "O4L32"
		SELECT CASE eng.chainReact
		CASE 1
			PLAY "G>C<L16G>E"
		CASE 2
			PLAY "G#>C#<L16G#>F"
		CASE 3
			PLAY "A>D<L16A>F#"
		CASE ELSE
			PLAY "A#>D#<L16A#>G"
		END SELECT
	CASE sndExplo2
		PLAY "O3L16C#<L32EF#FG"
	CASE sndExplo3
		PLAY "O0L64ED#"
	CASE sndThud
		PLAY "O0C10"
	CASE sndLevel
		PLAY "O3L32EL16G#EB>L10E"
	END SELECT
END SUB

''
'' TAKE A SCREENSHOT, SAVE TO BMP
''
SUB screenshot STATIC
	DIM cnt AS INTEGER, flName AS STRING * 12

	' initialize filename
	SELECT CASE cnt
	CASE 0
		cnt = 1
		flName = "CAPT0000.BMP"
	CASE IS > 9999
		EXIT SUB
	END SELECT

	' get next filename
	DO
		MID$(flName, 5, 4) = RIGHT$(STRING$(4, "0") + LTRIM$(STR$(cnt)), 4)
		cnt = cnt + 1
		IF (cnt > 9999) THEN EXIT SUB
	LOOP WHILE (fileExists%(flName))

	' write screen capture
	yCapture flName
END SUB

